# coding=utf-8
"""
© 2016. Case Recommender All Rights Reserved (License GPL3)

This file is responsible for write external files using data from tool.

The output file format is:
    user\titem\tinformation\n

* To change the spacing in the file, change the space_type var [default = \t].

Methods:
    - split_dataset: write train and test files for n folds and individuals feedback
    - cross_fold_validation: write train and test files for n folds.
    - write_prediction_file: write prediction file for rating prediction algorithms
    - write_ranking_file: write ranking file for item recommendation algorithms
    - write_ensemble: write ranking file for ensemble algorithms
"""

import os

__author__ = 'Arthur Fortes'


class WriteFile(object):
    def __init__(self, file_to_write, data, space_type="\t"):
        self.file_to_write = file_to_write
        self.data = data
        self.space_type = space_type

    @staticmethod
    def write(file_to_write, dict_structure, space, score=True):
        with open(file_to_write, 'w') as infile:
            for t in dict_structure:
                if score:
                    infile.write(str(t[0]) + space + str(t[1]) + space + str(t[2]) + '\n')
                else:
                    infile.write(str(t[0]) + space + str(t[1]) + space + str(1) + '\n')

    def split_dataset(self, feedback=None, dataset=None):
        names = list()

        if dataset is not None:
            for t in dataset:
                if t.rfind('/') == -1:
                    names.append(t[t.rfind('/')+1:])
                else:
                    names.append(t[t.rfind('/')+1:])

        self.file_to_write += "/folds/"

        if not os.path.exists(self.file_to_write):
            os.mkdir(self.file_to_write)

        for f in self.data:
            fold_name = self.file_to_write + str(f) + '/'
            if not os.path.exists(fold_name):
                os.mkdir(fold_name)

            self.write(fold_name + "train.dat", self.data[f]['train'], self.space_type)
            self.write(fold_name + "test.dat", self.data[f]['test'], self.space_type)

            if feedback is not None:
                for i, feed in enumerate(feedback[f]):
                    train_name = "train_" + str(names[i])
                    with open(fold_name + train_name, 'w') as infile:
                        for user in feed:
                            for item in feed[user]:
                                infile.write(str(user) + self.space_type + str(item) + self.space_type +
                                             str(feed[user][item]) + '\n')

    def cross_fold_validation(self):
        self.file_to_write += "/folds/"

        if not os.path.exists(self.file_to_write):
            os.mkdir(self.file_to_write)

        # f is an integer number (0 - n fold)
        for f in self.data:
            fold_name = self.file_to_write + str(f) + '/'
            if not os.path.exists(fold_name):
                os.mkdir(fold_name)

            self.write(fold_name + "train.dat", self.data[f]['train'], self.space_type)
            self.write(fold_name + "test.dat", self.data[f]['test'], self.space_type)

    def write_recommendation(self):
        self.write(self.file_to_write, self.data, self.space_type)

    def write_ensemble(self, list_users):
        with open(self.file_to_write, 'w') as infile:
            for user in list_users:
                for t in self.data[user]:
                    infile.write(str(user) + self.space_type + str(t[0]) + self.space_type + str(t[1]) + '\n')



from caserec.utils.cross_fold_validation import CrossFoldValidation

from caserec.evaluation.statistical_analysis import StatisticalAnalysis

from caserec.evaluation.rating_prediction import RatingPredictionEvaluation

from caserec.recommenders.rating_prediction.matrixfactorization import MatrixFactorization
from caserec.recommenders.rating_prediction.msmf import MSMF

# dir_dataset = "C:/Users/forte/OneDrive/msmf/datasets/filmtrust/"
# dir_dataset = "C:/Users/forte/OneDrive/msmf/datasets/ciaodvd/"
# dir_dataset = "C:/Users/forte/OneDrive/msmf/datasets/yahoo_movies/"
# dir_dataset = "C:/Users/forte/OneDrive/msmf/datasets/movielens/"
# dir_dataset = "C:/Users/forte/OneDrive/msmf/datasets/amazon/"
dir_dataset = "C:/Users/forte/OneDrive/msmf/datasets/booking_crossing/"

# CrossFoldValidation(input_file=dir_dataset + "user_rated_books_reduce.dat", dir_folds=dir_dataset).execute()


for f in range(10):
    fold_dir = dir_dataset + "folds/" + str(f) + "/"
    training_file = fold_dir + "train.dat"
    test_file = fold_dir + "test.dat"
    mf_file = fold_dir + "predict_mf.dat"
    msmf_file = fold_dir + "predict_msmf.dat"

    MatrixFactorization(training_file, test_file, prediction_file=mf_file, baseline=True).execute()
    MSMF(training_file, test_file, prediction_file=msmf_file, min_feedback=20, baseline=True,
         similarity_metric='correlation').execute()


# Evaluation

# msmf = RatingPredictionEvaluation().folds_evaluation(dir_dataset + 'folds/', n_folds=10,
#                                                      name_prediction="predict_msmf.dat", name_test="test.dat")
# mf = RatingPredictionEvaluation().folds_evaluation(dir_dataset + 'folds/', n_folds=10,
#                                                      name_prediction="predict_mf.dat", name_test="test.dat")


# msmf = RatingPredictionEvaluation().folds_evaluation_item_cold_start(dir_dataset + 'folds/', n_folds=10,
#                                                      name_prediction="predict_msmf.dat", name_test="test.dat", min_feedback=15)
# mf = RatingPredictionEvaluation().folds_evaluation_item_cold_start(dir_dataset + 'folds/', n_folds=10,
#                                                      name_prediction="predict_mf.dat", name_test="test.dat", min_feedback=15)

# print("RMSE MF:: ", mf[0], " RMSE MSMF::", msmf[0])
# StatisticalAnalysis(mf[2], msmf[2]).wilcoxon()

# print("MAE MF:: ", mf[3], " MAE MSMF::", msmf[3])
# StatisticalAnalysis(mf[5], msmf[5]).wilcoxon()



# coding=utf-8
"""
© 2017. Case Recommender All Rights Reserved (License GPL3)

This file contains item recommendation measures:
    MAP
    Precision
    Recall
    NDCG

    Class:

        - ItemRecommendationEvaluation

    Methods:

        - default_evaluation: returns:
            * prec@1, recall@1, ndcg@1, map@1 ... prec@10, recall@10, ndcg@10, map@10 and map_total
            * with only_map=True: map@1, map@3, map@5, map@10 and map_total
            * with only_ndcg=True: ndcg@1, ndcg@3, ndcg@5, ndcg@10

        - simple_evaluation: Evaluation only one ranking

        - all_but_one_evaluation: All-but-one Protocol

        - folds_evaluation: Evaluate N-Cross-Fold-Validation

        - evaluation_ranking: This method is used to build evaluate process into the recommenders files

"""

import numpy as np
from caserec.utils.extra_functions import check_error_file
from caserec.utils.read_file import ReadFile

__author__ = 'Arthur Fortes'


def ndcg_at_k(ranking):
    """
    Calculate the value of ndcg in ranking

    :param ranking: (numpy array) ranking
    :return: (float) value of ndcg
    """
    ranking = np.asfarray(ranking)
    r_ideal = np.asfarray(sorted(ranking, reverse=True))
    dcg_ideal = r_ideal[0] + np.sum(r_ideal[1:] / np.log2(np.arange(2, r_ideal.size + 1)))
    dcg_ranking = ranking[0] + np.sum(ranking[1:] / np.log2(np.arange(2, ranking.size + 1)))

    return dcg_ranking / dcg_ideal


class ItemRecommendationEvaluation(object):
    def __init__(self, space_type='\t', only_map=False, only_ndcg=False, n_ranks=list([1, 3, 5, 10])):
        """
        Class to evaluate rankings in a item recommendation scenario

        :param space_type: (string) delimiter (e.g ',' ' ' '\t')
        :param only_map: (bool) return only map values
        :param only_ndcg: (bool) return only ndcg values
        :param n_ranks: (list of int) list of positions to evaluate the ranking
        """
        self.space_type = space_type
        self.only_map = only_map
        self.only_ndcg = only_ndcg
        self.n_ranks = n_ranks

    def default_evaluation(self, ranking, test):
        """
        Calculate all the measures for item recommendation scenario

        :param ranking: (dict) dict generated by ReadFile(...).return_information()
        :param test:  (dict) dict generated by ReadFile(...).return_information()
        :return: (dict) values of the quality of ranking
        """

        avg_prec_total = list()
        final_values_dict = dict()
        num_user = len(test['users'])

        for i, n in enumerate(self.n_ranks):
            if n < 1:
                raise ValueError('Error: N must >= 1.')

            partial_precision = list()
            partial_recall = list()
            avg_prec_total = list()
            partial_ndcg = list()

            for user in test['users']:
                avg_prec_sum = 0

                hit_cont = 0
                # Generate user intersection list between the recommended items and test.
                list_feedback = set(list(ranking['du_order'].get(user, []))[:n])
                intersection = list(list_feedback.intersection(test['du_order'][user]))

                if len(intersection) > 0:
                    partial_precision.append((float(len(intersection)) / float(n)))
                    partial_recall.append((float(len(intersection)) / float(len(test['du'][user]))))

                    ig_ranking = np.zeros(n)
                    for item in intersection:
                        hit_cont += 1
                        avg_prec_sum += (float(hit_cont) / float(list(
                            ranking['du_order'][user])[:n].index(item) + 1))
                        ig_ranking[ranking['du_order'][user].index(item)] = 1

                    partial_ndcg.append(ndcg_at_k(ig_ranking))
                    avg_prec_total.append(float(avg_prec_sum) / float(len(test['du_order'][user])))

            # calculate final precision and recall
            if not self.only_map and not self.only_ndcg:
                final_values_dict.update({
                    'Prec@' + str(n): round(sum(partial_precision) / float(num_user), 5),
                    'Recall@' + str(n): round(sum(partial_recall) / float(num_user), 5),
                })

            if not self.only_map:
                final_values_dict.update({'NDCG@' + str(n): round(sum(partial_ndcg) / float(num_user), 5)})

            if not self.only_ndcg:
                final_values_dict.update({'MAP@' + str(n): round(sum(avg_prec_total) / float(num_user), 5)})

        if not self.only_ndcg:
            final_values_dict.update({'MAP': round(sum(avg_prec_total) / float(num_user), 5)})

        return final_values_dict

    def simple_evaluation(self, file_result, file_test):
        """
        A simple evaluation method to return the quality of a ranking

        :param file_result: (file) ranking file to evaluate
        :param file_test: (file) test file
        :return: Values of evaluation
        """

        # Verify that the files are valid
        check_error_file(file_result)
        check_error_file(file_test)

        predict = ReadFile(file_result, space_type=self.space_type).return_information()
        test = ReadFile(file_test, space_type=self.space_type).return_information()

        return self.default_evaluation(predict, test)

    def all_but_one_evaluation(self, file_result, file_test):
        """
        All-but-one Protocol: Considers only one pair (u, i) from the test set to evaluate the ranking

        :param file_result: (file) ranking file to evaluate
        :param file_test: (file) test file
        :return: Values of evaluation
        """

        # Verify that the files are valid
        check_error_file(file_result)
        check_error_file(file_test)

        predict = ReadFile(file_result, space_type=self.space_type).return_information()
        test = ReadFile(file_test, space_type=self.space_type).return_information()

        for user in test['users']:
            test['du'][user] = [list(test['du'][user])[0]]

        return self.default_evaluation(predict, test)

    def folds_evaluation(self, folds_dir, n_folds, name_prediction, name_test, type_recommendation='SimpleEvaluation',
                         no_deviation=False):
        """
        Evaluation N-Fold-Validation

        :param folds_dir: (string) fold of the files
        :param n_folds: (int) number of folds to evaluate
        :param name_prediction: (string) name of the ranking file [p.s: same in all folds]
        :param name_test: (string) name of the test file [p.s: same in all folds]
        :param type_recommendation: (string) choice 'SimpleEvaluation' or 'AllButOne'
        :param no_deviation: (bool) if true returns the standard deviation
        :return: Values of evaluation
        """
        result = list()
        list_results = list()

        for fold in range(n_folds):
            prediction = folds_dir + str(fold) + '/' + name_prediction
            test = folds_dir + str(fold) + '/' + name_test

            if type_recommendation == "SimpleEvaluation":
                result.append(self.simple_evaluation(prediction, test))
            elif type_recommendation == "AllButOne":
                result.append(self.all_but_one_evaluation(prediction, test))
            else:
                raise ValueError('Error: Invalid recommendation type!')

        for i in range(len(result[0])):
            list_partial = list()
            for j in range(n_folds):
                list_partial.append(result[j][i])
            if no_deviation:
                list_results.append(list_partial)
            else:
                list_results.append([np.mean(list_partial), np.std(list_partial)])

        return list_results

    def evaluation_ranking(self, ranking, test_file):
        ranking_dict = {'du_order': {}}
        test = ReadFile(test_file, space_type=self.space_type).return_information()

        for sample in ranking:
            ranking_dict['du_order'].setdefault(sample[0], list()).append(sample[1])

        return self.default_evaluation(ranking_dict, test)



# coding=utf-8
"""
© 2017. Case Recommender All Rights Reserved (License GPL3)

Matrix Factorization Based Collaborative Filtering Recommender

    Literature:
        Matrix Factorization Techniques for Recommender Systems
        http://dl.acm.org/citation.cfm?id=1608614

Parameters
-----------
    - train_file: string
    - test_file: string
    - prediction_file: string
        file to write final prediction
    - steps: int
         Number of steps over the training data
    - learn_rate: float
        Learning rate (alpha)
    - delta: float
        Regularization value
    - factors: int
        Number of latent factors per user/item
    - init_mean: float
        Mean of the normal distribution used to initialize the latent factors
    - init_stdev: float
        Standard deviation of the normal distribution used to initialize the latent factors
    - baseline: bool
        if True: Use the training data to build baselines (SVD Algorithm); else: Use only the mean
    - bias_learn_rate: float
        Learning rate for baselines
    - delta_bias: float
        Regularization value for baselines

"""

import numpy as np
from caserec.evaluation.rating_prediction import RatingPredictionEvaluation
from caserec.utils.extra_functions import timed
from caserec.utils.read_file import ReadFile
from caserec.utils.write_file import WriteFile

__author__ = "Arthur Fortes"


class MatrixFactorization(object):
    def __init__(self, train_file, test_file, prediction_file=None, steps=30, learn_rate=0.01, delta=0.015, factors=10,
                 init_mean=0.1, init_stdev=0.1, baseline=False, bias_learn_rate=0.005, delta_bias=0.002):
        self.train_set = ReadFile(train_file).return_information()
        self.test_set = ReadFile(test_file).return_information()
        self.prediction_file = prediction_file
        self.steps = steps
        self.learn_rate = learn_rate
        self.delta = delta
        self.factors = factors
        self.init_mean = init_mean
        self.init_stdev = init_stdev
        self.baseline = baseline
        self.predictions = list()
        self.map_items = dict()
        self.map_items_index = dict()
        self.map_users = dict()
        self.map_users_index = dict()
        self.bias_learn_rate = bias_learn_rate
        self.delta_bias = delta_bias

        self.p = None
        self.q = None
        self.bu = None
        self.bi = None

        self.users = sorted(set(list(self.train_set['users']) + list(self.test_set['users'])))
        self.items = sorted(set(list(self.train_set['items']) + list(self.test_set['items'])))

        for i, item in enumerate(self.items):
            self.map_items.update({item: i})
            self.map_items_index.update({i: item})
        for u, user in enumerate(self.users):
            self.map_users.update({user: u})
            self.map_users_index.update({u: user})

        list_feedback = list()
        self.dict_index = dict()
        for user, item, feedback in self.train_set['list_feedback']:
            list_feedback.append((self.map_users[user], self.map_items[item], feedback))
            self.dict_index.setdefault(self.map_users[user], []).append(self.map_items[item])
        self.train_set['list_feedback'] = list_feedback
        self._create_factors()

    def _create_factors(self):
        self.p = np.random.normal(self.init_mean, self.init_stdev, (len(self.users), self.factors))
        self.q = np.random.normal(self.init_mean, self.init_stdev, (len(self.items), self.factors))

        if self.baseline:
            self.bu = np.zeros(len(self.users), np.double)
            self.bi = np.zeros(len(self.items), np.double)

    def _predict(self, u, i, cond=True):
        if self.baseline:
            rui = self.train_set["mean_rates"] + self.bu[u] + self.bi[i] + np.dot(self.p[u], self.q[i])
        else:
            rui = self.train_set['mean_rates'] + np.dot(self.p[u], self.q[i])

        if cond:
            if rui > self.train_set["max"]:
                rui = self.train_set["max"]
            elif rui < self.train_set["min"]:
                rui = self.train_set["min"]

        return rui

    def train_model(self):
        rmse_old = .0
        for epoch in range(self.steps):
            error_final = .0
            for user, item, feedback in self.train_set['list_feedback']:
                eui = feedback - self._predict(user, item, False)
                error_final += (eui ** 2.0)

                # Adjust the factors
                u_f = self.p[user]
                i_f = self.q[item]

                # Compute factor updates
                delta_u = np.subtract(np.multiply(eui, i_f), np.multiply(self.delta, u_f))
                delta_i = np.subtract(np.multiply(eui, u_f), np.multiply(self.delta, i_f))

                # apply updates
                self.p[user] += np.multiply(self.learn_rate, delta_u)
                self.q[item] += np.multiply(self.learn_rate, delta_i)

                if self.baseline:
                    self.bu[user] += self.bias_learn_rate * (eui - self.delta_bias * self.bu[user])
                    self.bi[item] += self.bias_learn_rate * (eui - self.delta_bias * self.bi[item])

            rmse_new = np.sqrt(error_final / self.train_set["ni"])
            if np.fabs(rmse_new - rmse_old) <= 0.009:
                break
            else:
                rmse_old = rmse_new

    def predict(self):
        if self.test_set is not None:
            for user in self.test_set['users']:
                for item in self.test_set['feedback'][user]:
                    u, i = self.map_users[user], self.map_items[item]
                    self.predictions.append((user, item, self._predict(u, i, True)))

            if self.prediction_file is not None:
                self.predictions = sorted(self.predictions, key=lambda x: x[0])
                WriteFile(self.prediction_file, self.predictions).write_recommendation()
            return self.predictions

    def evaluate(self, predictions):
        result = RatingPredictionEvaluation()
        res = result.evaluation(predictions, self.test_set)
        print("Eval:: RMSE:" + str(res[0]) + " MAE:" + str(res[1]))

    def execute(self):
        # methods
        print("[Case Recommender: Rating Prediction > Matrix Factorization]\n")
        print("training data:: ", len(self.train_set['users']), " users and ", len(self.train_set['items']),
              " items and ", self.train_set['ni'], " interactions | sparsity ", self.train_set['sparsity'])
        print("test data:: ", len(self.test_set['users']), " users and ", len(self.test_set['items']),
              " items and ", (self.test_set['ni']), " interactions | sparsity ", self.test_set['sparsity'])
        print("training time:: ", timed(self.train_model), " sec")
        print("\nprediction_time:: ", timed(self.predict), " sec\n")
        self.evaluate(self.predictions)



# coding=utf-8
"""
© 2016. Case Recommender All Rights Reserved (License GPL3)

Item Based Collaborative Filtering Recommender

    Its philosophy is as follows: in order to determine the rating of User u on Movie m, we can find other movies that
    are similar to Movie m, and based on User u’s ratings on those similar movies we infer his rating on Movie m.

    Literature:
        http://cs229.stanford.edu/proj2008/Wen-RecommendationSystemBasedOnCollaborativeFiltering.pdf

Parameters
-----------
    - train_file: string
    - test_file: string
    - prediction_file: string
        file to write final prediction
    - similarity_metric: string
        Pairwise metric to compute the similarity between the users.
        Reference about distances:
            - http://docs.scipy.org/doc/scipy-0.14.0/reference/generated/scipy.spatial.distance.pdist.html
    - neighbors: int
        The number of item candidates strategy that you can choose for selecting the possible items to recommend.
    - space_type: string

"""

import numpy as np
from scipy.spatial.distance import squareform, pdist
from caserec.utils.extra_functions import timed
from caserec.utils.read_file import ReadFile
from caserec.recommenders.rating_prediction.base_knn import BaseKNNRecommenders
from caserec.utils.write_file import WriteFile

__author__ = 'Arthur Fortes'


class ItemKNN(BaseKNNRecommenders):
    def __init__(self, train_file, test_file, prediction_file=None, similarity_metric="correlation", neighbors=30,
                 space_type='\t'):
        self.train_set = ReadFile(train_file, space_type=space_type).return_information()
        self.test_set = ReadFile(test_file, space_type=space_type).return_information()
        BaseKNNRecommenders.__init__(self, self.train_set, self.test_set)
        self.k = neighbors
        self.prediction_file = prediction_file
        self.similarity_metric = similarity_metric
        self.predictions = list()
        self.si_matrix = None

    def compute_similarity(self):
        # Calculate distance matrix between items
        self.si_matrix = np.float32(squareform(pdist(self.matrix.T, self.similarity_metric)))
        # transform distances in similarities
        self.si_matrix = 1 - self.si_matrix
        del self.matrix

    '''
     for each pair (u,i) in test set, this method returns a prediction based
     on the others feedback in the train set

     rui = bui + (sum((ruj - buj) * sim(i,j)) / sum(sim(i,j)))

    '''
    def predict(self):
        if self.test is not None:
            for user in self.test['users']:
                for item_j in self.test['feedback'][user]:
                    list_n = list()
                    ruj = 0.0
                    try:
                        sum_sim = 0.0
                        for item_i in self.train['feedback'][user]:
                            try:
                                sim = self.si_matrix[self.map_items[item_i]][self.map_items[item_j]]
                                if np.math.isnan(sim):
                                    sim = 0.0
                            except KeyError:
                                sim = 0.0
                            list_n.append((item_i, sim))
                        list_n = sorted(list_n, key=lambda x: -x[1])

                        for pair in list_n[:self.k]:
                            try:
                                ruj += (self.train['feedback'][user][pair[0]] -
                                        self.bui[user][pair[0]]) * pair[1]
                                sum_sim += pair[1]
                            except KeyError:
                                pass

                        try:
                            ruj = self.bui[user][item_j] + (ruj / sum_sim)
                        except ZeroDivisionError:
                            ruj = self.bui[user][item_j]

                    except KeyError:
                        ruj = self.bui[user][item_j]

                    # normalize the ratings based on the highest and lowest value.
                    if ruj > self.train_set["max"]:
                        ruj = self.train_set["max"]
                    if ruj < self.train_set["min"]:
                        ruj = self.train_set["min"]
                    self.predictions.append((user, item_j, ruj))

            if self.prediction_file is not None:
                WriteFile(self.prediction_file, self.predictions).write_recommendation()
            return self.predictions

    def execute(self):
        # methods
        print("[Case Recommender: Rating Prediction > ItemKNN Algorithm]\n")
        print("training data:: ", len(self.train_set['users']), " users and ", len(self.train_set['items']),
              " items and ", self.train_set['ni'], " interactions | sparsity ", self.train_set['sparsity'])
        print("test data:: ", len(self.test_set['users']), " users and ", len(self.test_set['items']),
              " items and ", (self.test_set['ni']), " interactions | sparsity ", self.test_set['sparsity'])
        # training baselines bui
        self.fill_matrix()
        print("training time:: ", timed(self.train_baselines), " sec")
        self.compute_similarity()
        print("\nprediction_time:: ", timed(self.predict), " sec\n")
        self.evaluate(self.predictions)



# coding=utf-8
"""
© 2016. Case Recommender All Rights Reserved (License GPL3)

This file contains rating prediction measures:
    MAE
    RMSE
"""

import math
import numpy as np
from caserec.utils.read_file import ReadFile

__author__ = 'Arthur Fortes'


class RatingPredictionEvaluation(object):
    def __init__(self, space_type='\t'):
        self.space_type = space_type

    def simple_evaluation(self, file_result, file_test):
        predict = ReadFile(file_result, space_type=self.space_type).return_information()
        test = ReadFile(file_test, space_type=self.space_type).return_information()

        rmse = 0
        mae = 0
        count_comp = 0
        for user in test['users']:
            for item in test['feedback'][user]:
                try:
                    rui_predict = float(predict['feedback'][user][item])
                    rui_test = float(test['feedback'][user][item])
                    rmse += math.pow((rui_predict - rui_test), 2)
                    mae += math.fabs(rui_predict - rui_test)
                    count_comp += 1
                except KeyError:
                    pass

        if count_comp != 0:
            rmse = math.sqrt(float(rmse) / float(count_comp))
            mae = math.sqrt(float(mae) / float(count_comp))

        return rmse, mae

    @staticmethod
    def evaluation(predictions, test_set):
        rmse = 0
        mae = 0
        count_comp = 0

        for p in predictions:
            try:
                user, item, rui_predict = p[0], p[1], p[2]
                rui_test = float(test_set["feedback"][user][item])
                rmse += math.pow((rui_predict - rui_test), 2)
                mae += math.fabs(rui_predict - rui_test)
                count_comp += 1
            except KeyError:
                pass
        if count_comp != 0:
            rmse = math.sqrt(float(rmse) / float(count_comp))
            mae = math.sqrt(float(mae) / float(count_comp))

        return rmse, mae

    def folds_evaluation(self, folds_dir, n_folds, name_prediction, name_test):
        list_rmse = list()
        list_mae = list()
        
        for fold in range(n_folds):
            prediction = folds_dir + str(fold) + '/' + name_prediction
            test = folds_dir + str(fold) + '/' + name_test
            rmse, mae = self.simple_evaluation(prediction, test)
            list_rmse.append(rmse)
            list_mae.append(mae)

        return np.mean(list_rmse), np.std(list_rmse), list_rmse, np.mean(list_mae), np.std(list_mae), list_mae

    def folds_evaluation_item_cold_start(self, folds_dir, n_folds, name_prediction, name_test, min_feedback=10):
        list_rmse = list()
        list_mae = list()

        for fold in range(n_folds):
            train_file = folds_dir + str(fold) + '/' + 'train.dat'
            prediction = folds_dir + str(fold) + '/' + name_prediction
            test = folds_dir + str(fold) + '/' + name_test
            rmse, mae = self.simple_evaluation_item_cold_start(prediction, test, train_file, min_feedback=min_feedback)
            list_rmse.append(rmse)
            list_mae.append(mae)

        return np.mean(list_rmse), np.std(list_rmse), list_rmse, np.mean(list_mae), np.std(list_mae), list_mae

    def simple_evaluation_cold_start(self, file_result, file_test, list_user):
        predict = ReadFile(file_result, space_type=self.space_type).return_information()
        test = ReadFile(file_test, space_type=self.space_type).return_information()

        rmse = 0
        mae = 0
        count_comp = 0
        for user in test['users']:
            if user in list_user:
                for item in test['feedback'][user]:
                    try:
                        rui_predict = float(predict['feedback'][user][item])
                        rui_test = float(test['feedback'][user][item])
                        rmse += math.pow((rui_predict - rui_test), 2)
                        mae += math.fabs(rui_predict - rui_test)
                        count_comp += 1
                    except KeyError:
                        pass

        if count_comp != 0:
            rmse = math.sqrt(float(rmse) / float(count_comp))
            mae = math.sqrt(float(mae) / float(count_comp))

        return rmse, mae

    def simple_evaluation_item_cold_start(self, file_result, file_test, file_train, min_feedback=10):
        predict = ReadFile(file_result, space_type=self.space_type).return_information()
        test = ReadFile(file_test, space_type=self.space_type).return_information()
        train = ReadFile(file_train, space_type=self.space_type).return_information()
        new_items = set()
        for item in train['items']:
            if len(train['di'][item]) <= min_feedback:
                new_items.add(item)

        print(len(new_items))

        rmse = 0
        mae = 0
        count_comp = 0
        for user in test['users']:
                for item in test['feedback'][user]:
                    if item in new_items:
                        try:
                            rui_predict = float(predict['feedback'][user][item])
                            rui_test = float(test['feedback'][user][item])
                            rmse += math.pow((rui_predict - rui_test), 2)
                            mae += math.fabs(rui_predict - rui_test)
                            count_comp += 1
                        except KeyError:
                            pass

        if count_comp != 0:
            rmse = math.sqrt(float(rmse) / float(count_comp))
            mae = math.sqrt(float(mae) / float(count_comp))

        return rmse, mae



# coding=utf-8
"""
© 2016. Case Recommender All Rights Reserved (License GPL3)

This file has some auxiliary functions for Case Recommender. Method:
    - check_error_file: check if file exist
    - check_len_lists: check if the number of two list are equal
    - timed: measure the execution time of a function

"""

import sys
import time

__author__ = 'Arthur Fortes'


def check_error_file(file_check):
    try:
        open(file_check)
    except IOError:
        print("Error: File cannot be empty or file is invalid: " + str(file_check))
        sys.exit()


def check_len_lists(list1, list2):
    if len(list1) != len(list2):
        print("Error: Number of files in train list and rank list must be equal!")
        sys.exit()


def timed(f):
    start = time.time()
    f()
    elapsed = time.time() - start
    return elapsed



# coding=utf-8
"""
© 2017. Case Recommender All Rights Reserved (License GPL3)

SVD ++

Yehuda Koren: Factorization meets the neighborhood: a multifaceted collaborative filtering model, KDD 2008
http://portal.acm.org/citation.cfm?id=1401890.1401944

Parameters
-----------
    - train_file: string
    - test_file: string
    - prediction_file: string
        file to write final prediction
    - steps: int
         Number of steps over the training data
    - learn_rate: float
        Learning rate (alpha)
    - delta: float
        Regularization value
    - factors: int
        Number of latent factors per user/item
    - init_mean: float
        Mean of the normal distribution used to initialize the latent factors
    - init_stdev: float
        Standard deviation of the normal distribution used to initialize the latent factors
    - bias_learn_rate: float
        Learning rate for baselines
    - delta_bias: float
        Regularization value for baselines

"""

import numpy as np
from caserec.recommenders.rating_prediction.matrixfactorization import MatrixFactorization
from caserec.utils.extra_functions import timed
from caserec.utils.write_file import WriteFile

__author__ = "Arthur Fortes"


class SVDPlusPlus(MatrixFactorization):
    def __init__(self, train_file, test_file, prediction_file=None, steps=30, learn_rate=0.01, delta=0.015, factors=10,
                 init_mean=0.1, init_stdev=0.1, bias_learn_rate=0.005, bias_reg=0.002):
        MatrixFactorization.__init__(self, train_file=train_file, test_file=test_file, prediction_file=prediction_file,
                                     steps=steps, learn_rate=learn_rate, delta=delta, factors=factors,
                                     init_mean=init_mean, init_stdev=init_stdev, baseline=True,
                                     bias_learn_rate=bias_learn_rate, delta_bias=bias_reg)

        self.y = np.random.normal(self.init_mean, self.init_stdev, (len(self.items), self.factors))
        self.n_u = dict()
        # |N(u)|^(-1/2)
        for u, user in enumerate(self.users):
            self.n_u[u] = np.power(len(self.train_set["feedback"].get(user, [0])), -.5)

    def _predict_svd_plus_plus(self, u, i, sum_imp, cond=True):
        rui = self.train_set["mean_rates"] + self.bu[u] + self.bi[i] + np.dot((
            self.p[u] + self.n_u[u] * sum_imp), self.q[i])

        if cond:
            if rui > self.train_set["max"]:
                rui = self.train_set["max"]
            elif rui < self.train_set["min"]:
                rui = self.train_set["min"]
        return rui

    def train_model(self):
        rmse_old = .0
        for epoch in range(self.steps):
            error_final = .0
            for user, item, feedback in self.train_set['list_feedback']:

                # Incorporating implicit feedback in the SVD: Sum (j E N(u)) Yj
                sum_imp = sum(self.y[self.dict_index[user]])

                # Calculate error
                eui = feedback - self._predict_svd_plus_plus(user, item, sum_imp, False)
                error_final += (eui ** 2.0)

                # Adjust the factors
                u_f = self.p[user]
                i_f = self.q[item]

                # Compute factor updates
                delta_u = np.subtract(np.multiply(eui, i_f), np.multiply(self.delta, u_f))
                delta_i = np.subtract(np.multiply(eui, (u_f + self.n_u[user] * sum_imp)), np.multiply(self.delta, i_f))

                # apply updates
                self.p[user] += np.multiply(self.learn_rate, delta_u)
                self.q[item] += np.multiply(self.learn_rate, delta_i)

                # update bu and bi
                self.bu[user] += self.bias_learn_rate * (eui - self.delta_bias * self.bu[user])
                self.bi[item] += self.bias_learn_rate * (eui - self.delta_bias * self.bi[item])

                # update y (implicit factor)
                # ∀j E N(u) :
                # yj ← yj + γ2 · (eui · |N(u)|−1/2 · qi − λ7 · yj )

                for j in self.dict_index[user]:
                    delta_y = np.subtract(eui * self.n_u[user] * self.q[item], self.delta * self.y[j])
                    self.y[j] += self.learn_rate * delta_y

            rmse_new = np.sqrt(error_final / self.train_set["ni"])

            if np.fabs(rmse_new - rmse_old) <= 0.009:
                break
            else:
                rmse_old = rmse_new

    def predict(self):
        if self.test_set is not None:
            for user in self.test_set['users']:

                # sum (j E N(u)) Yj
                try:
                    sum_imp = sum(self.y[self.dict_index[user]])
                except KeyError:
                    sum_imp = np.ones(self.factors, np.double)

                for item in self.test_set['feedback'][user]:
                    self.predictions.append((user, item, self._predict_svd_plus_plus(
                        self.map_users[user], self.map_items[item], sum_imp)))

            if self.prediction_file is not None:
                self.predictions = sorted(self.predictions, key=lambda x: x[0])
                WriteFile(self.prediction_file, self.predictions).write_recommendation()
            return self.predictions

    def execute(self):
        # methods
        print("[Case Recommender: Rating Prediction > SVD++]\n")
        print("training data:: ", len(self.train_set['users']), " users and ", len(self.train_set['items']),
              " items and ", self.train_set['ni'], " interactions | sparsity ", self.train_set['sparsity'])
        print("test data:: ", len(self.test_set['users']), " users and ", len(self.test_set['items']),
              " items and ", (self.test_set['ni']), " interactions | sparsity ", self.test_set['sparsity'])
        print("training time:: ", timed(self.train_model), " sec")
        print("\nprediction_time:: ", timed(self.predict), " sec\n")
        self.evaluate(self.predictions)



# coding=utf-8
"""
© 2016. Case Recommender All Rights Reserved (License GPL3)

NSVD1
    Literature:
    Recommending new movies: even a few ratings are more valuable than metadata
    http://dl.acm.org/citation.cfm?id=1639731


Parameters
-----------
    - train_file: string
    - test_file: string
    - metadata_file: string
        Metadata file ; Format file:
        item \t metadata \t value\n
     - prediction_file: string
     - steps: int
        Number of steps over the training data
     - learn_rate: float
        Learning rate
     - delta: float
        Regularization value
     - factors: int
        Number of latent factors per user/item
     - init_mean: float
        Mean of the normal distribution used to initialize the latent factors
     - init_stdev: float
        Standard deviation of the normal distribution used to initialize the latent factors
     - alpha: float
     - batch: bool
        if True: Use batch model to train the model (default False)
     - n2: int
        Number of interactions in batch step
     - learn_rate2: float
        Learning rate in batch step
     - delta2: float
        Regularization value in Batch step

"""

from caserec.recommenders.rating_prediction.base_nsvd1 import BaseNSVD1
from caserec.utils.extra_functions import timed
from caserec.utils.read_file import ReadFile
import numpy as np
import math

__author__ = "Arthur Fortes"


class ItemNSVD1(BaseNSVD1):
    def __init__(self, train_file, test_file, metadata_file, prediction_file=None, steps=30, learn_rate=0.01,
                 delta=0.015, factors=10, init_mean=0.1, init_stdev=0.1, alpha=0.001, batch=False, n2=10,
                 learn_rate2=0.01, delta2=0.015, space_type='\t'):
        BaseNSVD1.__init__(self, train_file, test_file, prediction_file, factors, init_mean, init_stdev, space_type)
        self.metadata = ReadFile(metadata_file, space_type=space_type).read_metadata(self.items)
        self.number_metadata = len(self.metadata["metadata"])
        self.batch = batch
        self.steps = steps
        self.learn_rate = learn_rate
        self.delta = delta
        self.alpha = alpha
        self.n2 = n2
        self.learn_rate2 = learn_rate2
        self.delta2 = delta2

        # Internal
        self.x = self.metadata['matrix']
        self.non_zero_x = list()
        self.d = list()
        for i in range(self.number_items):
            self.non_zero_x.append(list(np.where(self.x[i] != 0)[0]))
            with np.errstate(divide='ignore'):
                self.d.append(1 / np.dot(self.x[i].T, self.x[i]))

    def _update_factors(self, item, i):
        c, e = 0, 0
        try:
            for user in self.train_set['di'][item]:
                u = self.map_users[user]
                rui = self._predict(u, i)
                error = self.train_set['feedback'][user][item] - rui
                b = np.array(self.p[u])

                # update factors
                self.p[u] += self.learn_rate * (error * self.q[i] - self.delta * self.p[u])
                self.q[i] += self.learn_rate * (error * b - self.delta * self.q[i])
                self.b[u] += self.learn_rate * (error - self.delta * self.b[u])
                self.c[i] += self.learn_rate * (error - self.delta * self.c[i])
                c += 1
                e += error ** 2
        except KeyError:
            pass
        return c, e

    def train_model(self):
        for k in range(self.steps):
            rmse = 0
            count_error = 0
            for i, item in enumerate(self.items):
                self.q[i] = np.dot(self.x[i], self.w)
                a = np.array(self.q[i])
                c, e = self._update_factors(item, i)
                rmse += e
                count_error += c

                for l in self.non_zero_x[i]:
                    self.w[l] += self.d[i] * self.x[i][l] * (self.q[i] - a)
            rmse = math.sqrt(rmse / float(count_error))

            if (math.fabs(rmse - self.last_rmse)) <= self.alpha:
                break
            else:
                self.last_rmse = rmse
            print("step::", k, "RMSE::", rmse)

    def train_batch_model(self):
        for k in range(self.steps):
            rmse = 0
            count_error = 0
            self.q = np.dot(self.x, self.w)

            for i, item in enumerate(self.items):
                c, e = self._update_factors(item, i)
                rmse += e
                count_error += c

            for _ in range(self.n2):
                for i, item in enumerate(self.items):
                    e = self.q[i] - (np.dot(self.x[i], self.w))

                    for l in self.non_zero_x[i]:
                        self.w[l] += self.learn_rate2 * (self.d[i] * np.dot(self.x[i][l], e.T) -
                                                         (self.w[l] * self.delta2))

            self.q = np.dot(self.x, self.w)

            rmse = math.sqrt(rmse / float(count_error))
            if (math.fabs(rmse - self.last_rmse)) <= self.alpha:
                break
            else:
                self.last_rmse = rmse
            print("step::", k, "RMSE::", rmse)

    def execute(self):
        # methods
        print("[Case Recommender: Rating Prediction > Item NSVD1]\n")
        print("training data:: ", len(self.train_set['users']), " users and ", len(self.train_set['items']),
              " items and ", self.train_set['ni'], " interactions | sparsity ", self.train_set['sparsity'])
        print("test data:: ", len(self.test_set['users']), " users and ", len(self.test_set['items']),
              " items and ", self.test_set['ni'], " interactions | sparsity ", self.test_set['sparsity'])
        print("metadata:: ", len(self.metadata['items']), " items and ", len(self.metadata['metadata']),
              " metadata and ", self.metadata['ni'], " interactions\n")
        self._create_factors()

        if self.batch:
            print("training time:: ", timed(self.train_batch_model)), " sec\n"
        else:
            print("training time:: ", timed(self.train_model)), " sec\n"

        print("\nprediction_time:: ", timed(self.predict)), " sec\n"
        self.evaluate(self.predictions)



# coding=utf-8
"""
© 2016. Case Recommender All Rights Reserved (License GPL3)

k-fold cross-validation

    In k-fold cross-validation, the original sample is randomly partitioned into k equal sized subsamples.
    Of the k subsamples, a single subsample is retained as the validation data for testing the model, and the remaining
    k − 1 subsamples are used as training data. The cross-validation process is then repeated k times (the folds),
    with each of the k subsamples used exactly once as the validation data.

    The k results from the folds can then be averaged (or otherwise combined) to produce a single estimation.
    The advantage of this method over repeated random sub-sampling (see below) is that all observations are used for
     both training and validation, and each observation is used for validation exactly once. 10-fold cross-validation
     is commonly used, but in general k remains an unfixed parameter.

    When k=n (the number of observations), the k-fold cross-validation is exactly the leave-one-out cross-validation.

    In stratified k-fold cross-validation, the folds are selected so that the mean response value is approximately equal
    in all the folds. In the case of a dichotomous classification, this means that each fold contains roughly the
    same proportions of the two types of class labels.

    Literature:
        https://en.wikipedia.org/wiki/Cross-validation_(statistics)

Parameters
-----------
    - input_file: string
    - space_type: string
        TDivide data space type. E.g:
            '\t'  = tabular [default]
            ','   = comma
            '.'   = dot
    - dir_folds: string
        Directory to write the folds
    - n_folds: int
        Number to divide data in folds

"""

import random

from caserec.utils.extra_functions import timed
from caserec.utils.read_file import ReadFile
from caserec.utils.write_file import WriteFile

__author__ = 'Arthur Fortes'


class CrossFoldValidation(object):
    def __init__(self, input_file, space_type='\t', dir_folds=None, n_folds=10):
        self.dataset = input_file
        self.n_folds = n_folds
        self.space_type = space_type
        self.dict_folds = dict()
        self.dir_folds = dir_folds
        self.tp = None

    def divide_dataset(self):
        self.tp = ReadFile(self.dataset, space_type=self.space_type).return_information()
        random.shuffle(self.tp['list_feedback'])

        # Get the number of interactions that each partition should have.
        partition_size = int(float(self.tp['ni']) / float(self.n_folds))

        list_folds = list()
        last = -1

        for p in range(self.n_folds):
            initial = 1 + last
            final = (p + 1) * partition_size
            list_folds.append(self.tp['list_feedback'][initial:final])
            last = final

        for fold in range(self.n_folds):
            train_set = list()
            for fold_train in range(self.n_folds):
                if fold_train != fold:
                    train_set += list_folds[fold_train]
                train_set.sort()

            list_folds[fold].sort()
            self.dict_folds[fold] = {'train': train_set, 'test': list_folds[fold]}

        if self.dir_folds is not None:
            WriteFile(self.dir_folds, self.dict_folds, self.space_type).cross_fold_validation()

    def execute(self):
        print("[Case Recommender: Cross-Fold-Validation Process]\n")
        print("number of folds:: ", self.n_folds, " write folds in:: ", self.dir_folds)
        print("total process time:: ", timed(self.divide_dataset), "sec")
        print("dataset info:: #users: ", len(self.tp['users']), " #items: ", len(self.tp['items']),
              " #number of interactions: ", self.tp['ni'], " #range of feedback: [",
              self.tp['min'], "-", self.tp['max'], "]\n")



# coding=utf-8
"""
© 2016. Case Recommender All Rights Reserved (License GPL3)

Given a dataset, this function is responsible for creating a set of training and test.

Standard file for reading:

user /t item /t feedback /n

* To change the spacing in the file, change the space_type var [default = \t].

By default algorithm divides the base into 90% for training and 10% for test.
To modify this parameter, change "test_ratio".


Parameters
-----------
    - dataset: list of dir files
        if you want to insert only one file: ['/dir_fold/file.dat']
    - dir_folds: string
        dir path to write folds
    - n_folds: int
    - test_ratio
        value to divide to test, e.g 0.1 = 10%
    - space_type string
    
"""

import random

from caserec.utils.extra_functions import timed
from caserec.utils.read_file import ReadFile
from caserec.utils.write_file import WriteFile

__author__ = 'Arthur Fortes'


class SplitDataset(object):
    def __init__(self, dataset, dir_folds=None, n_folds=10, test_ratio=0.1, space_type='\t'):
        self.dataset = dataset
        self.n_folds = n_folds
        self.space_type = space_type
        self.test_ratio = test_ratio
        self.dict_folds = dict()
        self.dict_feedback_folds = dict()
        self.dir_folds = dir_folds

    def divide_dataset(self):
        tp = ReadFile(self.dataset, space_type=self.space_type)
        tp.split_dataset()

        for fold in range(self.n_folds):
            dict_feedback = list()
            tp.triple_dataset = list(set(tp.triple_dataset))
            random.shuffle(tp.triple_dataset)
            sp = int((1-self.test_ratio) * len(tp.triple_dataset))
            train = tp.triple_dataset[:sp]
            test = tp.triple_dataset[sp:]
            train.sort()
            test.sort(key=lambda x: x[0])
            train_set = list()
            test_set = list()

            for i, feedback in enumerate(self.dataset):
                dict_individual = dict()

                for triple in train:
                    try:
                        dict_individual.setdefault(triple[0], {}).update(
                            {triple[1]: tp.individual_interaction[i][triple[0]][triple[1]]})
                        train_set.append([triple[0], triple[1], tp.individual_interaction[i][triple[0]][triple[1]]])
                    except KeyError:
                        pass

                for triple_test in test:
                    try:
                        test_set.append([triple_test[0], triple_test[1],
                                         tp.individual_interaction[i][triple_test[0]][triple_test[1]]])
                    except KeyError:
                        pass

                dict_feedback.append(dict_individual)

            self.dict_feedback_folds[fold] = dict_feedback
            self.dict_folds[fold] = {'train': train_set, 'test': test_set}

        if self.dir_folds is not None:
            WriteFile(self.dir_folds, self.dict_folds, self.space_type).split_dataset(self.dict_feedback_folds,
                                                                                      self.dataset)

    def execute(self):
        print("[Case Recommender: Split Dataset Process]\n")
        print("number of folds:: ", self.n_folds, " write folds in:: ", self.dir_folds)
        print("total process time:: ", timed(self.divide_dataset), "sec")



# coding=utf-8
"""
© 2016. Case Recommender All Rights Reserved (License GPL3)

This file is responsible for read external files.

The accepted data format is:
    user\titem\tinformation\n

* user and item must be integer!
* To change the spacing in the file, change the space_type var (default = \t).

Methods:
    - main_information and main_information_item_recommendation: returns a set of information about the dataset:
            * list with all users
            * list with all items
            * number of interactions
            * users interactions (dictionary with seen items and feedback for each user |
                                  number of interaction for each user)
            * items interactions (dictionary with users and feedback for each item |
                                  number of interaction for each item)

    - cross_fold_validation: return triples [user, item, feedback] and number of interactions

    - split_dataset: return triples [user, item, feedback], number of interactions and
                       users interactions (dictionary with seen items and feedback for each user |
                                  number of interaction for each user) for each feedback type.

    - rating_prediction: returns a set of specifics attributes from dataset in a dictionary:
            * dictionary with all interactions
            * list with all users
            * list with all items
            * dictionary of all users interaction
            * dictionary of all items interaction
            * mean of rates
    - read_rankings: return a dictionary and a list about one ranking
    - read_matrix: returns a data matrix

"""

import sys
import copy
import numpy as np
from caserec.utils.extra_functions import check_error_file

__author__ = 'Arthur Fortes'


class ReadFile(object):
    def __init__(self, file_read, space_type='\t'):
        self.file_read = file_read
        self.space_type = space_type
        self.list_users = set()
        self.list_items = set()
        self.number_interactions = 0
        self.dict_users = dict()
        self.dict_items = dict()
        self.num_user_interactions = dict()
        self.num_items_interactions = dict()
        self.triple_dataset = list()
        self.individual_interaction = list()
        self.average_scores = dict()
        self.mean_feedback = 0

    def return_information(self, implicit=False):
        check_error_file(self.file_read)
        dict_file = dict()
        d_feedback = dict()
        list_feedback = list()
        not_seen = dict()
        map_user = dict()
        map_index_user = dict()
        du_feed = dict()

        with open(self.file_read) as infile:
            for line in infile:
                if line.strip():
                    inline = line.split(self.space_type)
                    if len(inline) == 1:
                        print("Error: Space type is invalid!")
                        sys.exit()
                    try:
                        user, item, feedback = int(inline[0]), int(inline[1]), float(inline[2])
                        d_feedback.setdefault(user, {}).update({item: feedback})
                        self.triple_dataset.append((user, item, feedback))
                        self.dict_users.setdefault(user, set()).add(item)
                        self.dict_items.setdefault(item, set()).add(user)
                        du_feed.setdefault(user, list()).append(item)
                        self.list_users.add(user)
                        self.list_items.add(item)
                        self.mean_feedback += feedback
                        list_feedback.append(feedback)
                        self.number_interactions += 1
                    except ValueError:
                        pass

        self.triple_dataset = sorted(self.triple_dataset)
        self.mean_feedback /= float(self.number_interactions)
        self.list_users = set(sorted(list(self.list_users)))
        self.list_items = set(sorted(list(self.list_items)))

        for user in self.list_users:
            not_seen[user] = list(set(self.list_items) - set(self.dict_users[user]))

        for u, user in enumerate(self.list_users):
            map_user[user] = u
            map_index_user[u] = user

        map_item = dict()
        map_index_item = dict()
        self.list_items = set(sorted(list(self.list_items)))

        for i, item in enumerate(self.list_items):
            map_item[item] = i
            map_index_item[i] = item

        matrix = np.zeros((len(self.list_users), len(self.list_items)))

        di = copy.deepcopy(self.dict_items)

        for user in self.list_users:
            for item in self.dict_users[user]:
                if implicit:
                    matrix[map_user[user]][map_item[item]] = 1
                else:
                    matrix[map_user[user]][map_item[item]] = d_feedback[user][item]
                self.dict_items.setdefault(map_item[item], set()).add(map_user[user])

        sparsity = (1 - (self.number_interactions / float(len(self.list_users) * len(self.list_items)))) * 100
        dict_file.update({'feedback': d_feedback, 'users': self.list_users, 'items': self.list_items,
                          'du': self.dict_users, 'dir': self.dict_items, 'mean_rates': self.mean_feedback,
                          'list_feedback': self.triple_dataset, 'ni': self.number_interactions,
                          'max': max(list_feedback), 'min': min(list_feedback), 'sparsity': sparsity,
                          'not_seen': not_seen, 'matrix': matrix, 'map_user': map_index_user,
                          'map_item': map_index_item, 'mu': map_user, 'mi': map_item, 'du_order': du_feed, 'di': di})

        return dict_file

    def triple_information(self):
        check_error_file(self.file_read)
        with open(self.file_read) as infile:
            for line in infile:
                if line.strip():
                    inline = line.split(self.space_type)
                    if len(inline) == 1:
                        print("Error: Space type is invalid!")
                        sys.exit()

                    try:
                        user, item, feedback = int(inline[0]), int(inline[1]), inline[2].replace("\n", "")
                        self.triple_dataset.append([user, item, feedback])
                        self.number_interactions += 1
                    except ValueError:
                        pass

    def split_dataset(self):
        for i, feedback in enumerate(self.file_read):
            self.dict_users = dict()
            check_error_file(feedback)
            with open(feedback) as infile:
                for line in infile:
                    if line.strip():
                        inline = line.split(self.space_type)
                        if len(inline) == 1:
                            print("Error: Space type is invalid!")
                            sys.exit()

                        self.number_interactions += 1
                        try:
                            user, item, feedback = int(inline[0]), int(inline[1]), float(inline[2])
                            self.triple_dataset.append((user, item))
                            self.dict_users.setdefault(user, {}).update({item: feedback})
                        except ValueError:
                            pass

            self.individual_interaction.append(self.dict_users)

    def read_rankings(self):
        list_feedback = list()
        check_error_file(self.file_read)
        with open(self.file_read) as infile:
            for line in infile:
                if line.strip():
                    inline = line.split(self.space_type)
                    if len(inline) == 1:
                        print("Error: Space type is invalid!")
                        sys.exit()
                    try:
                        user, item, feedback = int(inline[0]), int(inline[1]), float(inline[2])
                        self.dict_users.setdefault(user, {}).update({item: feedback})
                        list_feedback.append(feedback)
                        self.average_scores[user] = self.average_scores.get(user, 0) + feedback
                        self.num_user_interactions[user] = self.num_user_interactions.get(user, 0) + 1
                    except ValueError:
                        pass

        return self.dict_users, list_feedback

    def read_matrix(self):
        matrix = list()
        check_error_file(self.file_read)
        with open(self.file_read) as infile:
            for line in infile:
                if line.strip():
                    inline = line.split(self.space_type)
                    inline = np.array(inline)
                    inline = np.delete(inline, len(inline)-1)
                    matrix.append(inline.astype(float))
        return np.array(matrix)

    def ensemble(self):
        dict_info = dict()

        for r, rank_file in enumerate(self.file_read):
            self.list_users = set()
            self.dict_users = dict()
            with open(rank_file) as infile:
                for line in infile:
                    if line.strip():
                        inline = line.split(self.space_type)
                        if len(inline) == 1:
                            print("Error: Space type is invalid!")
                            sys.exit()

                        try:
                            user, item, score = int(inline[0]), int(inline[1]), float(inline[2])
                            self.list_users.add(user)
                            self.dict_users.setdefault(user, list()).append([user, item, score])
                        except ValueError:
                            pass

            self.list_users = set(sorted(self.list_users))
            for user in self.list_users:
                n_rank = len(self.dict_users[user])
                for i, triple in enumerate(self.dict_users[user]):
                    self.dict_users[user][i][2] = n_rank - i

            dict_info.setdefault(r, dict()).update({'rank': self.dict_users, 'users': self.list_users})

        return dict_info

    def ensemble_test(self):
        user_info = dict()
        rank_info = dict()
        for r, rank_file in enumerate(self.file_read):
            r_dict = dict()
            with open(rank_file) as infile:
                for line in infile:
                    if line.strip():
                        inline = line.split(self.space_type)
                        if len(inline) == 1:
                            print("Error: Space type is invalid!")
                            sys.exit()
                        try:
                            user, item, score = int(inline[0]), int(inline[1]), float(inline[2])
                            self.number_interactions += 1
                            self.list_users.add(user)
                            self.list_items.add(item)
                            self.dict_users.setdefault(user, {}).update({item: score})
                            r_dict.setdefault(user, {}).update({item: score})
                        except ValueError:
                            pass
            rank_info[r] = r_dict

        self.list_users = set(sorted(self.list_users))
        self.list_items = set(sorted(self.list_items))
        dict_non_seen = dict()

        for user in self.dict_users:
            dict_non_seen[user] = list(set(self.list_items) - set(self.dict_users[user]))
            user_info[user] = {'j': dict_non_seen[user], 'i': self.dict_users[user].keys()}

        return user_info, self.list_users, self.list_items, self.number_interactions, rank_info

    def read_metadata(self, l_items):
        dict_file = dict()
        d_feedback = dict()
        list_feedback = list()
        map_user = dict()
        map_index_user = dict()
        map_item = dict()
        map_index_item = dict()
        check_error_file(self.file_read)

        with open(self.file_read) as infile:
            for line in infile:
                if line.strip():
                    inline = line.split(self.space_type)
                    if len(inline) == 1:
                        print("Error: Space type is invalid in metadata file!")
                        print(inline, self.space_type)
                        sys.exit()
                    try:
                        user, item, feedback = int(inline[0]), int(inline[1]), float(inline[2])
                        d_feedback.setdefault(user, {}).update({item: feedback})
                        self.triple_dataset.append((user, item, feedback))
                        self.dict_users.setdefault(user, set()).add(item)
                        self.dict_items.setdefault(item, set()).add(user)
                        self.list_items.add(item)
                        self.mean_feedback += feedback
                        list_feedback.append(feedback)
                        self.number_interactions += 1
                    except ValueError:
                        pass

        self.triple_dataset = sorted(self.triple_dataset)
        self.mean_feedback /= float(self.number_interactions)
        self.list_users = set(sorted(list(l_items)))
        self.list_items = set(sorted(list(self.list_items)))

        for u, user in enumerate(self.list_users):
            map_user[user] = u
            map_index_user[u] = user

        for i, item in enumerate(self.list_items):
            map_item[item] = i
            map_index_item[i] = item

        matrix = np.zeros((len(self.list_users), len(self.list_items)))

        for user in self.list_users:
            try:
                for item in d_feedback[user]:
                    matrix[map_user[user]][map_item[item]] = d_feedback[user][item]
            except KeyError:
                pass

        dict_file.update({'feedback': d_feedback, 'items': self.list_users, 'metadata': self.list_items,
                          'di': self.dict_users, 'dm': self.dict_items, 'mean_rates': self.mean_feedback,
                          'list_feedback': self.triple_dataset, 'ni': self.number_interactions,
                          'max': max(list_feedback), 'min': min(list_feedback), 'matrix': matrix})

        return dict_file



# coding=utf-8
"""
© 2016. Case Recommender All Rights Reserved (License GPL3)

This file is base for neighborhood-base algorithms

Parameters
-----------
    - train_set: dict
        train data dictionary generated for ReadFile(file).rating_prediction()
    - train_set: dict
        test data dictionary generated for ReadFile(file).rating_prediction()
"""

import numpy as np
from caserec.evaluation.rating_prediction import RatingPredictionEvaluation

__author__ = 'Arthur Fortes'


class BaseKNNRecommenders(object):
    def __init__(self, train_set, test_set):
        self.train = train_set
        self.test = test_set
        self.regBi = 10
        self.regBu = 15
        self.bu = dict()
        self.bi = dict()
        self.bui = dict()
        self.users = sorted(set(list(self.train['users']) + list(self.test['users'])))
        self.items = sorted(set(list(self.train['items']) + list(self.test['items'])))
        self.number_users = len(self.users)
        self.number_items = len(self.items)
        self.map_items = dict()
        self.map_users = dict()
        self.matrix = None

        for item_id, item in enumerate(self.items):
            self.map_items[item] = item_id

        for user_id, user in enumerate(self.users):
            self.map_users[user] = user_id

    def fill_matrix(self):
        self.matrix = np.zeros((self.number_users, self.number_items))
        for u, user in enumerate(self.users):
            try:
                for item in self.train['feedback'][user]:
                    self.matrix[u][self.map_items[item]] = self.train['feedback'][user][item]
            except KeyError:
                pass

    def train_baselines(self):
        for i in range(10):
            self.compute_bi()
            self.compute_bu()
        self.compute_bui()

    def compute_bi(self):
        # bi = (rui - mi - bu) / (regBi + number of interactions)
        self.bi = dict()

        for item in self.train['items']:
            cont = 0
            # self.bi.update({item: 0})
            for user in self.train['di'][item]:
                self.bi[item] = self.bi.get(item, 0) + float(self.train['feedback'][user].get(item, 0)) - \
                                self.train['mean_rates'] - self.bu.get(user, 0)
                cont += 1
            if cont > 1:
                self.bi[item] = float(self.bi[item]) / float(self.regBi + cont)

    def compute_bu(self):
        # bu = (rui - mi - bi) / (regBu + number of interactions)
        self.bu = dict()
        for user in self.train['users']:
            cont = 0
            for item in self.train['du'][user]:
                self.bu[user] = self.bu.get(user, 0) + float(self.train['feedback'][user].get(item, 0)) - \
                                self.train['mean_rates'] - self.bi.get(item, 0)
                cont += 1
            if cont > 1:
                self.bu[user] = float(self.bu[user]) / float(self.regBu + cont)

    def compute_bui(self):
        # bui = mi + bu + bi
        for user in self.users:
            for item in self.items:
                try:
                    self.bui.setdefault(user, {}).update(
                        {item: self.train['mean_rates'] + self.bu[user] + self.bi[item]})
                except KeyError:
                    self.bui.setdefault(user, {}).update({item: self.train['mean_rates']})
        del self.bu
        del self.bi

    def evaluate(self, predictions):
        result = RatingPredictionEvaluation()
        res = result.evaluation(predictions, self.test)
        print("\nEval:: RMSE:", res[0], " MAE:", res[1])



# coding=utf-8
"""
© 2016. Case Recommender All Rights Reserved (License GPL3)

This file contains Statical functions for recommender systems.
    - T-test
    - Wilcoxon
"""

from scipy.stats import ttest_ind, ranksums
import numpy as np

__author__ = 'Arthur Fortes'


class StatisticalAnalysis(object):
    def __init__(self, sample1, sample2, method='ttest'):
        self.sample1 = np.array(sample1)
        self.sample2 = np.array(sample2)
        self.method = method

    def general_analysis(self):
        """
        Analyzing the difference

        Instead you might compute the difference and apply some common measure like the sum of absolute differences 
        (SAD), the sum of squared differences (SSD) or the correlation coefficient:
        """

        print("=== Information About Samples ===")
        print("Standard Deviation Sample1: " + str(np.std(self.sample1)))
        print("Standard Deviation Sample2: " + str(np.std(self.sample2)) + "\n")
        print("=== Analyzing the Difference Between Samples ===")
        print("SAD:" + str(np.sum(np.abs(self.sample1 - self.sample2))))
        print("SSD:" + str(np.sum(np.square(self.sample1 - self.sample2))))
        print("Correlation:" + str(np.corrcoef(np.array((self.sample1, self.sample2)))[0, 1]) + "\n")

    def ttest(self):
        """
        T-student
    
        Calculates the T-test for the means of TWO INDEPENDENT samples of scores.
    
        This is a two-sided test for the null hypothesis that 2 independent samples have identical 
        average (expected) values
    
        This test assumes that the populations have identical variances.
        """

        t, p = ttest_ind(self.sample1, self.sample2)
        print("=== T- Student Analysis ===")
        print("The calculated t-statistic: " + str(t))
        print("The two-tailed p-value: " + str(p) + "\n")

    def wilcoxon(self):
        """
        Wilcoxon
        
        The Wilcoxon signed-rank test tests the null hypothesis that two related paired samples come from 
        the same distribution. In particular, it tests whether the distribution of the differences x - y 
        is symmetric about zero. It is a non-parametric version of the paired T-test.
        """

        t, p = ranksums(self.sample1, self.sample2)
        print("=== Wilcoxon Analysis ===")
        print("The calculated t-statistic: " + str(t))
        print("The two-tailed p-value: " + str(p) + "\n")

    def execute(self):
        self.general_analysis()
        if self.method.lower() == "wilcoxon":
            self.wilcoxon()
        elif self.method.lower() == "ttest":
            self.ttest()
        else:
            print("Error: Method Invalid!")



# coding=utf-8
"""
© 2016. Case Recommender All Rights Reserved (License GPL3)

Item Based Collaborative Filtering Recommender with Attributes

    Its philosophy is as follows: in order to determine the rating of User u on Movie m, we can find other movies that
    are similar to Movie m, and based on User u’s ratings on those similar movies we infer his rating on Movie m.
    However, instead of traditional ItemKNN, this approach uses a pre-computed similarity matrix.

    Literature:
        http://cs229.stanford.edu/proj2008/Wen-RecommendationSystemBasedOnCollaborativeFiltering.pdf

Parameters
-----------
    - train_file: string
    - test_file: string
    - prediction_file: string
        file to write final prediction
    - metadata_file: string
        Metadata file ; Format file:
        item \t metadata \t value\n
    - similarity_matrix_file: string
        Pairwise metric to compute the similarity between the users based on a set of attributes.
        Format file:
        Distances separated by \t, where the users should be ordering. E g.:
        distance1\tdistance2\tdistance3\n
        distance1\tdistance2\tdistance3\n
        distance1\tdistance2\tdistance3\n
    - neighbors: int
        The number of item candidates strategy that you can choose for selecting the possible items to recommend.
    - similarity_metric: string
    - space_type: string

"""

import sys
from caserec.recommenders.rating_prediction.itemknn import ItemKNN
from caserec.utils.extra_functions import timed
from caserec.utils.read_file import ReadFile

__author__ = 'Arthur Fortes'


class ItemAttributeKNN(ItemKNN):
    def __init__(self, train_file, test_file, metadata_file=None, similarity_matrix_file=None, prediction_file=None,
                 neighbors=30, similarity_metric="correlation", space_type='\t'):
        ItemKNN.__init__(self, train_file, test_file, prediction_file=prediction_file, neighbors=neighbors,
                         similarity_metric=similarity_metric, space_type=space_type)

        if metadata_file is None and similarity_matrix_file is None:
            print("This algorithm needs a similarity matrix or a metadata file!")
            sys.exit(0)

        if metadata_file is not None:
            self.metadata = ReadFile(metadata_file, space_type=space_type).read_metadata(self.items)
            self.matrix = self.metadata['matrix'].T
        self.similarity_matrix_file = similarity_matrix_file

    def read_matrix(self):
        self.si_matrix = ReadFile(self.similarity_matrix_file).read_matrix()

    def execute(self):
        # methods
        print("[Case Recommender: Rating Prediction > Item Attribute KNN Algorithm]\n")
        print("training data:: ", len(self.train_set['users']), " users and ", len(self.train_set['items']),
              " items and ", self.train_set['ni'], " interactions | sparsity ", self.train_set['sparsity'])
        print("test data:: ", len(self.test_set['users']), " users and ", len(self.test_set['items']),
              " items and ", (self.test_set['ni']), " interactions | sparsity ", self.test_set['sparsity'])

        # training baselines bui
        print("training time:: ", timed(self.train_baselines), " sec")
        if self.similarity_matrix_file is not None:
            print("compute similarity:: ", timed(self.read_matrix), " sec")
        else:
            print("compute similarity time:: ", timed(self.compute_similarity), " sec")
        print("\nprediction_time:: ", timed(self.predict)), " sec\n"
        self.evaluate(self.predictions)



# coding=utf-8
"""
© 2016. Case Recommender All Rights Reserved (License GPL3)

Base NSVD1

Parameters
-----------
    - train_file: string
    - test_file: string
     - prediction_file: string
     - steps: int
        Number of steps over the training data
     - learn_rate: float
        Learning rate
     - delta: float
        Regularization value
     - factors: int
        Number of latent factors per user/item
     - init_mean: float
        Mean of the normal distribution used to initialize the latent factors
     - init_stdev: float
        Standard deviation of the normal distribution used to initialize the latent factors
     - alpha: float
     - batch: bool
        if True: Use batch model to train the model (default False)
     - n2: int
        Number of interactions in batch step
     - learn_rate2: float
        Learning rate in batch step
     - delta2: float
        Regularization value in Batch step

"""

from caserec.evaluation.rating_prediction import RatingPredictionEvaluation
from caserec.utils.read_file import ReadFile
from caserec.utils.write_file import WriteFile
import numpy as np

__author__ = "Arthur Fortes"


class BaseNSVD1(object):
    def __init__(self, train_file, test_file, prediction_file=None, factors=10, init_mean=0.1, init_stdev=0.1,
                 space_type='\t'):
        self.train_set = ReadFile(train_file, space_type=space_type).return_information()
        self.test_set = ReadFile(test_file, space_type=space_type).return_information()
        self.prediction_file = prediction_file
        self.factors = factors
        self.init_mean = init_mean
        self.init_stdev = init_stdev
        self.users = sorted(set(list(self.train_set["users"]) + list(self.test_set["users"])))
        self.items = sorted(set(list(self.train_set["items"]) + list(self.test_set["items"])))
        self.number_users = len(self.users)
        self.number_items = len(self.items)
        self.metadata = None
        self.number_metadata = None
        self.map_items = dict()
        self.map_items_index = dict()
        self.map_users = dict()
        self.map_users_index = dict()
        for i, item in enumerate(self.items):
            self.map_items.update({item: i})
            self.map_items_index.update({i: item})
        for u, user in enumerate(self.users):
            self.map_users.update({user: u})
            self.map_users_index.update({u: user})

        # internal vars
        self.x = None
        self.p = None
        self.q = None
        self.w = None
        self.b = None
        self.c = None
        self.last_rmse = 0
        self.predictions = list()

    def _create_factors(self):
        self.b = self.init_mean * np.random.randn(self.number_users, 1) + self.init_stdev ** 2
        self.c = self.init_mean * np.random.randn(self.number_items, 1) + self.init_stdev ** 2
        self.p = self.init_mean * np.random.randn(self.number_users, self.factors) + self.init_stdev ** 2
        self.q = self.init_mean * np.random.randn(self.number_items, self.factors) + self.init_stdev ** 2
        self.w = self.init_mean * np.random.randn(self.number_metadata, self.factors) + self.init_stdev ** 2

    def _predict(self, user, item):
        rui = self.b[user] + self.c[item] + np.dot(self.p[user], self.q[item])
        return rui[0]

    def predict(self):
        if self.test_set is not None:
            for user in self.test_set['users']:
                for item in self.test_set['feedback'][user]:
                    try:
                        rui = self._predict(self.map_users[user], self.map_items[item])
                        if rui > self.train_set["max"]:
                            rui = self.train_set["max"]
                        if rui < self.train_set["min"]:
                            rui = self.train_set["min"]
                        self.predictions.append((user, item, rui))
                    except KeyError:
                        self.predictions.append((user, item, self.train_set["mean_rates"]))

            if self.prediction_file is not None:
                WriteFile(self.prediction_file, self.predictions).write_recommendation()
            return self.predictions

    def evaluate(self, predictions):
        result = RatingPredictionEvaluation()
        res = result.evaluation(predictions, self.test_set)
        print("\nEval:: RMSE:", res[0], " MAE:", res[1], '\n')



"""
© 2017. Case Recommender All Rights Reserved (License GPL3)

Item Most Similar Matrix Factorization - Item MSMF

"""

import numpy as np
from scipy.spatial.distance import squareform, pdist
from caserec.recommenders.rating_prediction.matrixfactorization import MatrixFactorization
from caserec.utils.extra_functions import timed
from caserec.utils.write_file import WriteFile

__author__ = "Arthur Fortes"


class MSMF(MatrixFactorization):
    def __init__(self, train_file, test_file, prediction_file=None, steps=30, learn_rate=0.01, delta=0.015, factors=10,
                 init_mean=0.1, init_stdev=0.1, baseline=False, similarity_metric='cosine', min_feedback=10):

        MatrixFactorization.__init__(self, train_file=train_file, test_file=test_file, steps=steps,
                                     learn_rate=learn_rate, delta=delta, factors=factors, init_mean=init_mean,
                                     init_stdev=init_stdev, baseline=baseline)
        self.prediction_file = prediction_file
        self.similarity_metric = similarity_metric
        self.min_feedback = min_feedback
        self.si_matrix = None

    def compute_similarity(self):
        # Calculate distance matrix between items
        self.si_matrix = np.float32(squareform(pdist(self.train_set['matrix'].T, self.similarity_metric)))
        # transform distances in similarities
        self.si_matrix = 1 - self.si_matrix

    def search_similar_item(self, item):
        item_index = self.train_set['mi'][item]
        similar_list = sorted(range(len(self.si_matrix[item_index])), key=lambda k: -self.si_matrix[item_index][k])
        for i in similar_list:
            if i != item_index and len(self.train_set['di'][self.train_set['map_item'][i]]) > self.min_feedback:
                return self.train_set['map_item'][i]

    def predict_test(self):
        cont_new_item = set()

        if self.test_set is not None:
            for user in self.test_set['users']:
                for item in self.test_set['feedback'][user]:
                    try:
                        # check if item is new in the dataset
                        if len(self.train_set['di'][item]) < self.min_feedback:
                            cont_new_item.add(item)
                            # if user is new, the algorithm find the user most similar with its and replace
                            item = self.search_similar_item(item)
                    except KeyError:
                        pass

                    u, i = self.map_users[user], self.map_items[item]
                    self.predictions.append((user, item, self._predict(u, i, True)))

            if self.prediction_file is not None:
                self.predictions = sorted(self.predictions, key=lambda x: x[0])
                WriteFile(self.prediction_file, self.predictions).write_recommendation()
            print("Cont num cold-start items:: " + str(len(cont_new_item)))
            return self.predictions

    def execute(self):
        # methods
        print("[Case Recommender: Rating Prediction > Item MSMF]\n")
        print("training data:: ", len(self.train_set['users']), " users and ", len(self.train_set['items']),
              " items and ", self.train_set['ni'], " interactions | sparsity ", self.train_set['sparsity'])
        print("test data:: ", len(self.test_set['users']), " users and ", len(self.test_set['items']),
              " items and ", (self.test_set['ni']), " interactions | sparsity ", self.test_set['sparsity'])

        # compute similarity
        self.compute_similarity()
        # train model
        print("training time:: ", timed(self.train_model), " sec")
        print("\nprediction_time:: ", timed(self.predict_test), " sec\n")
        self.evaluate(self.predictions)


dir_fold = "C:/Users/forte/OneDrive/msmf/datasets/filmtrust/folds/0/"
# MatrixFactorization(dir_fold + "train.dat", dir_fold + "test.dat").execute()
MSMF(dir_fold + "train.dat", dir_fold + "test.dat", min_feedback=10, similarity_metric='cosine').execute()



# coding=utf-8
"""
© 2016. Case Recommender All Rights Reserved (License GPL3)

Ensemble BPR Learning - Optimizing ensemble rankings

    Literature:
        Arthur Fortes da Costa and Marcelo G. Manzato
        Exploiting multimodal interactions in recommender systems with ensemble algorithms
        Journal Information Systems 2016.
        http://www.sciencedirect.com/science/article/pii/S0306437915300818

Parameters
-----------

"""

import numpy as np

from caserec.evaluation.item_recommendation import ItemRecommendationEvaluation
from caserec.utils.read_file import ReadFile
from caserec.utils.write_file import WriteFile

__author__ = 'Arthur Fortes'


class EnsembleLearningBPR(object):
    def __init__(self, list_train_files, list_rankings_files, ranking_file=None, test_file=None, rank_number=10,
                 space_type='\t', init_mean=0.1, init_stdev=0.1, factors=10, num_interactions=30, learn_rate=0.05,
                 reg_u=0.0025, reg_i=0.0025, reg_j=0.00025, reg_bias=0, use_loss=True):
        self.list_train_files = list_train_files
        self.list_rankings_files = list_rankings_files
        self.ranking_file = ranking_file
        self.test_file = test_file
        self.rank_number = rank_number
        self.space_type = space_type
        self.init_mean = init_mean
        self.init_stdev = init_stdev
        self.factors = factors
        self.num_interactions = num_interactions
        self.learn_rate = learn_rate
        self.reg_bias = reg_bias
        self.reg_u = reg_u
        self.reg_i = reg_i
        self.reg_j = reg_j
        self.use_loss = use_loss

        # external vars
        self.train_info = None
        self.rankings_info = None
        self.list_users = None
        self.list_items = None
        self.number_users = None
        self.number_items = None
        self.num_events = None
        self.ir = None
        self.rf = dict()
        self.map_user = dict()
        self.map_item = dict()
        self.loss = None
        self.loss_sample = list()
        self.ranking = list()

    def _create_factors(self):
        p = self.init_mean * np.random.randn(self.number_users, self.factors) + self.init_stdev ** 2
        q = self.init_mean * np.random.randn(self.number_items, self.factors) + self.init_stdev ** 2
        bias = self.init_mean * np.random.randn(self.number_items, 1) + self.init_stdev ** 2
        beta = self.init_mean * np.random.randn(self.number_users, 1) + self.init_stdev ** 2
        return p, q, bias, beta

    def read_training_data(self):
        self.train_info, self.list_users, self.list_items, self.num_events, _ = ReadFile(
            self.list_train_files).ensemble_test()
        self.number_users = len(self.list_users)
        self.number_items = len(self.list_items)
        # remove

        self.num_events = 5000

        for u, user in enumerate(self.list_users):
            self.map_user[user] = u
        for i, item in enumerate(self.list_items):
            self.map_item[item] = i

        for r in range(len(self.list_train_files)):
            p, q, bias, beta = self._create_factors()
            self.rf[r] = {"p": p, "q": q, "bias": bias, "beta": beta}

    def _compute_loss(self):
        ranking_loss = 0
        for sample in self.loss_sample:
            x_uij = self._predict_score(sample[0], sample[1]) - self._predict_score(sample[0], sample[2])
            ranking_loss += 1 / (1 + np.exp(x_uij))

        complexity = 0
        for r in range(len(self.list_train_files)):
            for sample in self.loss_sample:
                u, i, j = sample[0], sample[1], sample[2]
                u, i, j = self.map_user[u], self.map_item[i], self.map_item[j]
                complexity += self.reg_u * np.power(np.linalg.norm(self.rf[r]["p"][u]), 2)
                complexity += self.reg_i * np.power(np.linalg.norm(self.rf[r]["q"][i]), 2)
                complexity += self.reg_j * np.power(np.linalg.norm(self.rf[r]["q"][j]), 2)
                complexity += self.reg_bias * np.power(self.rf[r]["bias"][i], 2)
                complexity += self.reg_bias * np.power(self.rf[r]["bias"][j], 2)

        return (ranking_loss/2.0) + 0.5 * (complexity/2.0)

    def read_rankings(self):
        self.rankings_info, _, _, _, self.ir = ReadFile(
            self.list_rankings_files).ensemble_test()

    def _sample_triple(self):
        u = np.random.choice(self.list_users)
        i = np.random.choice(self.train_info[u]["i"])
        j = np.random.choice(self.rankings_info[u]["i"])
        return u, i, j

    def _predict_score(self, user, item):
        u, i = self.map_user[user], self.map_item[item]
        rating = 0
        for r in range(len(self.list_train_files)):
            rating += self.rf[r]["bias"][i] + np.dot(self.rf[r]["p"][u], self.rf[r]["q"][i])
        return rating

    def _update_factors(self, user, item_i, item_j):
        u, i, j = self.map_user[user], self.map_item[item_i], self.map_item[item_j]
        rui = 0
        ruj = 0
        x_uij = 0

        for r in range(len(self.list_rankings_files)):
            try:
                self.ir[r][user][item_i] = 0
                rui += self.rf[r]["bias"][i] + np.dot(self.rf[r]["p"][u], self.rf[r]["q"][i])
                ruj += self.rf[r]["bias"][j] + np.dot(self.rf[r]["p"][u], self.rf[r]["q"][j])
            except KeyError:
                pass

            # x_uij += self.rf[r]["beta"][u] * (rui - ruj)
            x_uij += (rui - ruj)

        eps = 1 / (1 + np.exp(x_uij))

        for r in range(len(self.list_rankings_files)):
            try:
                self.ir[r][user][item_i] = 0
                self.rf[r]["bias"][i] += self.learn_rate * (eps - self.reg_bias * self.rf[r]["bias"][i])
                self.rf[r]["bias"][j] += self.learn_rate * (eps - self.reg_bias * self.rf[r]["bias"][j])

                # Adjust the factors
                u_f = self.rf[r]["p"][u]
                i_f = self.rf[r]["q"][i]
                j_f = self.rf[r]["q"][j]

                # Compute factor updates
                delta_u = (i_f - j_f) * eps - self.reg_u * u_f
                delta_i = u_f * eps - self.reg_i * i_f
                delta_j = -u_f * eps - self.reg_j * j_f

                self.rf[r]["p"][u] += self.learn_rate * delta_u
                self.rf[r]["q"][i] += self.learn_rate * delta_i
                self.rf[r]["q"][j] += self.learn_rate * delta_j

            except KeyError:
                pass

    def train_model(self):
        if self.use_loss:
            num_sample_triples = int(np.sqrt(len(self.map_user)) * 100)
            for _ in range(num_sample_triples):
                self.loss_sample.append(self._sample_triple())
            self.loss = self._compute_loss()

        for i in range(self.num_interactions):
            for j in range(self.num_events):
                user, item_i, item_j = self._sample_triple()
                self._update_factors(user, item_i, item_j)

            if self.use_loss:
                actual_loss = self._compute_loss()
                if actual_loss > self.loss:
                    self.learn_rate *= 0.5
                elif actual_loss < self.loss:
                    self.learn_rate *= 1.1
                self.loss = actual_loss

    def predict(self):
        for user in self.rankings_info:
            partial_ranking = list()
            for item in self.rankings_info[user]["i"]:
                partial_ranking.append((self.map_user[user], self.map_item[item], self._predict_score(user, item)))
            partial_ranking = sorted(partial_ranking, key=lambda x: -x[2])[:self.rank_number]
            self.ranking += partial_ranking

        if self.ranking_file is not None:
            WriteFile(self.ranking_file, self.ranking).write_recommendation()

    def evaluate(self, measures):
        res = ItemRecommendationEvaluation().evaluation_ranking(self.ranking, self.test_file)
        evaluation = 'Eval:: '
        for measure in measures:
            evaluation += measure + ': ' + str(res[measure]) + ' '
        print(evaluation)

    def execute(self, measures=('Prec@5', 'Prec@10', 'NDCG@5', 'NDCG@10', 'MAP@5', 'MAP@10')):
        # methods
        print("[Case Recommender: Item Recommendation > Ensemble BPR Learning Algorithm]\n")
        self.read_training_data()
        self.read_rankings()
        self.train_model()
        self.predict()
        if self.test_file is not None:
            self.evaluate(measures)



# coding=utf-8
"""
© 2016. Case Recommender All Rights Reserved (License GPL3)

User Based Collaborative Filtering Recommender

    User-kNN predicts a user’s rating according to how similar users rated the same item. The algorithm matches similar
    users based on the similarity of their ratings on items.


Parameters
-----------
    - train_file: string
    - test_file: string
    - ranking_file: string
        file to write final ranking
    - similarity_metric: string
        Pairwise metric to compute the similarity between the users.
        Reference about distances:
            - http://docs.scipy.org/doc/scipy-0.14.0/reference/generated/scipy.spatial.distance.pdist.html
    - neighbors: int
        The number of user candidates strategy that you can choose for selecting the possible items to recommend.
    - rank_number int
        The number of items per user that appear in final rank
    - implicit bool
        If True define fill matrix with 0s and 1s
    - space_type: string

"""

import numpy as np
from scipy.spatial.distance import squareform, pdist
from caserec.evaluation.item_recommendation import ItemRecommendationEvaluation
from caserec.utils.extra_functions import timed
from caserec.utils.read_file import ReadFile
from caserec.utils.write_file import WriteFile

__author__ = "Arthur Fortes"


class UserKNN(object):
    def __init__(self, train_file, test_file=None, ranking_file=None, similarity_metric="correlation", neighbors=30,
                 rank_number=10, implicit=False, space_type='\t'):
        self.train_set = ReadFile(train_file, space_type=space_type).return_information(implicit)
        self.test_file = test_file
        self.users = self.train_set['users']
        self.items = self.train_set['items']
        if self.test_file is not None:
            self.test_set = ReadFile(test_file).return_information()
            self.users = sorted(list(self.train_set['users']) + list(self.test_set['users']))
            self.items = sorted(list(self.train_set['items']) + list(self.test_set['items']))
        self.k = neighbors
        self.similarity_metric = similarity_metric
        self.ranking_file = ranking_file
        self.rank_number = rank_number
        self.space_type = space_type
        self.ranking = list()
        self.su_matrix = None
        self.matrix = self.train_set['matrix']

    def compute_similarity(self):
        # Calculate distance matrix between users
        self.su_matrix = np.float32(squareform(pdist(self.matrix, self.similarity_metric)))
        # transform distances in similarities
        self.su_matrix = 1 - self.su_matrix

    def _predict_score(self, user_i, user_j):
        pass

    def predict(self):
        for user in range(len(self.train_set['matrix'])):
            partial_ranking = list()
            u_list = list(np.flatnonzero(self.train_set['matrix'][user] == 0))
            neighbors = sorted(range(len(self.su_matrix[user])), key=lambda m: -self.su_matrix[user][m])

            for item in u_list:
                sim_suv = 0
                common_user_neighbor = list(set(self.train_set['dir'][item]).intersection(set(neighbors[1:self.k])))

                if len(common_user_neighbor) > 0:
                    for user_neighbor in common_user_neighbor:
                        sim = 0 if np.math.isnan(self.su_matrix[user][user_neighbor]
                                                 ) else self.su_matrix[user][user_neighbor]
                        sim_suv += sim

                    partial_ranking.append((self.train_set['map_user'][user], self.train_set['map_item'][item],
                                            sim_suv))

            partial_ranking = sorted(partial_ranking, key=lambda x: -x[2])[:self.rank_number]
            self.ranking += partial_ranking

        if self.ranking_file is not None:
            WriteFile(self.ranking_file, self.ranking).write_recommendation()

    def evaluate(self, measures):
        res = ItemRecommendationEvaluation().evaluation_ranking(self.ranking, self.test_file)
        evaluation = 'Eval:: '
        for measure in measures:
            evaluation += measure + ': ' + str(res[measure]) + ' '
        print(evaluation)

    def execute(self, measures=('Prec@5', 'Prec@10', 'NDCG@5', 'NDCG@10', 'MAP@5', 'MAP@10')):
        print("[Case Recommender: Item Recommendation > UserKNN Algorithm]\n")
        print("training data:: ", len(self.train_set['users']), " users and ", len(self.train_set['items']),
              " items and ", self.train_set['ni'], " interactions | sparsity ", self.train_set['sparsity'])

        if self.test_file is not None:
            test_set = ReadFile(self.test_file, space_type=self.space_type).return_information()
            print("test data:: ", len(self.test_set['users']), " users and ", len(self.test_set['items']),
                  " items and ", (self.test_set['ni']), " interactions | sparsity ", self.test_set['sparsity'])
            del test_set
        print("training time:: ", timed(self.compute_similarity), " sec")
        print("prediction_time:: ", timed(self.predict), " sec\n")
        if self.test_file is not None:
            self.evaluate(measures)



# coding=utf-8
"""
© 2016. Case Recommender All Rights Reserved (License GPL3)

This file has some functions to build some structures by attributes or training data. Methods:
    - read_dataset -> Simple Dictionary:: user:{item: feedback}
    - read_attribute -> Dictionary whit list:: category:[list of items]
    - compute_write_similarity_matrix -> compute similarity_matrix and write the result after

Parameters
-----------
    - input_file: string
    - space_type: string
        Divide data space type. E.g:
            '\t'  = tabular [default]
            ','   = comma
            '.'   = dot

"""

import numpy as np
from scipy.spatial.distance import squareform, pdist

__author__ = 'Arthur Fortes'


class TreatData(object):
    def __init__(self, input_file, type_space="\t"):
        self.input_file = input_file
        self.type_space = type_space
        self.item_category_dict = dict()
        self.dataset_dict = dict()
        self.category_list = set()

    def read_dataset(self):
        with open(self.input_file) as infile:
            for line in infile:
                if line.strip():
                    inline = line.split(self.type_space)
                    user, item, feedback = inline[0], inline[1], inline[2]
                    self.dataset_dict.setdefault(user, {}).update({item: feedback})
    '''
    - category_file: string
        file that has categories and them respective items
    '''
    def read_attribute(self, category_file):
        with open(category_file) as infile:
            for line in infile:
                if line.strip():
                    inline = line.split(self.type_space)
                    inline[1] = inline[1].replace('\n', '')
                    self.category_list.add(inline[1])
                    self.item_category_dict.setdefault(inline[0], []).append(inline[1])
    '''
     - distance: string
        Pairwise metric to compute the similarity between the users/ items.
     - write_file: string
        file to write similarity matrix
     - user_matrix: bool
        if True build a matrix user x user; else build a matrix item x item
    '''
    def compute_write_distance_matrix(self, distance, write_file, user_matrix=True):
        lu = set()
        li = set()

        with open(self.input_file) as infile:
            for line in infile:
                if line.strip():
                    inline = line.split(self.type_space)
                    user, item, feedback = int(inline[0]), int(inline[1]), float(inline[2])
                    self.dataset_dict.setdefault(user, {}).update({item: feedback})
                    lu.add(user)
                    li.add(item)

        matrix = np.zeros((len(lu), len(li)))
        lu = sorted(list(lu))
        li = sorted(list(li))

        map_items = dict()
        map_users = dict()
        for i, item in enumerate(li):
            map_items.update({item: i})
        for u, user in enumerate(lu):
            map_users.update({user: u})

        for user in self.dataset_dict:
            for item in self.dataset_dict[user]:
                matrix[map_users[user]][map_items[item]] = self.dataset_dict[user][item]

        if not user_matrix:
            matrix = matrix.T

        matrix = np.float32(squareform(pdist(matrix, distance)))
        matrix -= 1
        with open(write_file, "w") as infile:
            for i in range(len(matrix)):
                for j in range(len(matrix[0])):
                    infile.write(str(matrix[i][j]) + self.type_space)
                infile.write("\n")



from caserec.utils.read_file import ReadFile
from caserec.utils.write_file import WriteFile


class BaseEnsemble(object):
    def __init__(self, list_ranks, test_file, write_file=""):
        self.list_ranks = list_ranks
        self.test_file = test_file
        self.write_file = write_file
        self.dict_ranks = ReadFile(self.list_ranks).ensemble()
        self.list_users = set()
        self.final_ranking = dict()

        # methods
        self.ensemble()
        self.write_results()

    def ensemble(self):
        for rank in self.dict_ranks:
            for user in self.dict_ranks[rank]["users"]:
                self.list_users.add(user)
                for i, item in enumerate(self.dict_ranks[rank]["rank"][user]):
                    try:
                        self.final_ranking[item[0]][item[1]] += item[2]
                    except KeyError:
                        self.final_ranking.setdefault(user, {}).update({item[1]: item[2]})
        self.list_users = set(sorted(self.list_users))
        for user in self.list_users:
            self.final_ranking[user] = sorted(self.final_ranking[user].items(), key=lambda x: -x[1])[:10]

    def write_results(self):
        WriteFile(self.write_file, self.final_ranking).write_ensemble(self.list_users)



# coding=utf-8
"""
© 2017. Case Recommender All Rights Reserved (License GPL3)

Most Popular Collaborative Filtering Recommender


Parameters
-----------
    - train_file: string
    - test_file: string
    - ranking_file: string
        file to write final ranking
    - rank_number int
        The number of items per user that appear in final rank
    - space_type: string

"""

from caserec.evaluation.item_recommendation import ItemRecommendationEvaluation
from caserec.utils.extra_functions import timed
from caserec.utils.read_file import ReadFile
from caserec.utils.write_file import WriteFile


class MostPopular(object):
    def __init__(self, train_file, test_file=None, ranking_file=None, rank_number=10, space_type='\t'):
        self.train_set = ReadFile(train_file, space_type=space_type).return_information()
        self.test_file = test_file
        self.users = self.train_set['users']
        self.items = self.train_set['items']
        if self.test_file is not None:
            self.test_set = ReadFile(test_file).return_information()
            self.users = sorted(list(self.train_set['users']) + list(self.test_set['users']))
            self.items = sorted(list(self.train_set['items']) + list(self.test_set['items']))
        self.ranking_file = ranking_file
        self.rank_number = rank_number
        self.space_type = space_type
        self.ranking = list()

    def predict(self):
        for user in set(self.users):
            rank_user = list()
            for item in self.train_set['not_seen'].get(user, []):
                ni = len(self.train_set['di'][item])
                nu = len(self.train_set['du'][user])
                rank_user.append((user, item, round(ni/float(nu), 4)))
            rank_user = sorted(rank_user, key=lambda x: -x[2])
            self.ranking += rank_user[:self.rank_number]

        if self.ranking_file is not None:
            WriteFile(self.ranking_file, self.ranking).write_recommendation()

    def evaluate(self, measures):
        res = ItemRecommendationEvaluation().evaluation_ranking(self.ranking, self.test_file)
        evaluation = 'Eval:: '
        for measure in measures:
            evaluation += measure + ': ' + str(res[measure]) + ' '
        print(evaluation)

    def execute(self, measures=('Prec@5', 'Prec@10', 'NDCG@5', 'NDCG@10', 'MAP@5', 'MAP@10')):
        print("[Case Recommender: Item Recommendation > Most Popular Algorithm]\n")
        print("training data:: ", len(self.train_set['users']), " users and ", len(self.train_set['items']),
              " items and ", self.train_set['ni'], " interactions | sparsity ", self.train_set['sparsity'])

        if self.test_file is not None:
            test_set = ReadFile(self.test_file, space_type=self.space_type).return_information()
            print("test data:: ", len(self.test_set['users']), " users and ", len(self.test_set['items']),
                  " items and ", (self.test_set['ni']), " interactions | sparsity ", self.test_set['sparsity'])
            del test_set
        print("prediction_time:: ", timed(self.predict), " sec\n")
        if self.test_file is not None:
            self.evaluate(measures)



from distutils.core import setup
from setuptools import find_packages

__author__ = "Arthur Fortes"

setup(
    name='CaseRecommender',
    packages=find_packages(),
    version='0.0.18',
    license='GPL3 License',
    description='A recommender systems framework for Python',
    author='Arthur Fortes',
    author_email='fortes.arthur@gmail.com',
    url='https://github.com/ArthurFortes/CaseRecommender',
    download_url='https://github.com/ArthurFortes/CaseRecommender/tarball/0.0.18',
    keywords=['recommender systems', 'framework', 'collaborative filtering', 'content-based filtering'],
    classifiers=[],
)



# coding=utf-8
"""
© 2016. Case Recommender All Rights Reserved (License GPL3)

Matrix factorization model for item prediction (ranking) optimized using BPR.

 * BPR reduces ranking to pairwise classification.
    Literature:
        Steffen Rendle, Christoph Freudenthaler, Zeno Gantner, Lars Schmidt-Thieme:
        BPR: Bayesian Personalized Ranking from Implicit Feedback.
        UAI 2009.
        http://www.ismll.uni-hildesheim.de/pub/pdfs/Rendle_et_al2009-Bayesian_Personalized_Ranking.pdf

Parameters
-----------
    - train_file: string
    - test_file: string
    - ranking_file: string
        file to write final ranking
    - factors: int
        Number of latent factors per user/item
    - learn_rate: float
        Learning rate (alpha)
    - num_interactions: int
        Number of iterations over the training data
    - num_events: int
        Number of events in each interaction
        -> default: None -> number of interactions of train file
    - predict_items_number: int
        Number of items per user in ranking
    - init_mean: float
        Mean of the normal distribution used to initialize the latent factors
    - init_stdev: float
        Standard deviation of the normal distribution used to initialize the latent factors
    - reg_u: float
        Regularization parameter for user factors
    - reg_i: float
        Regularization parameter for positive item factors
    - reg_j: float
        Regularization parameter for negative item factors
    - reg_bias: float
        Regularization parameter for the bias term
    - use_loss: bool
        Use objective function to increase learning rate
    - space_type: string

"""

import numpy as np
from caserec.evaluation.item_recommendation import ItemRecommendationEvaluation
from caserec.utils.extra_functions import timed
from caserec.utils.read_file import ReadFile
from caserec.utils.write_file import WriteFile

__author__ = "Arthur Fortes"


class BprMF(object):
    def __init__(self, train_file, test_file=None, ranking_file=None, factors=10, learn_rate=0.05, num_interactions=30,
                 num_events=None, predict_items_number=10, init_mean=0.1, init_stdev=0.1, reg_u=0.0025, reg_i=0.0025,
                 reg_j=0.00025, reg_bias=0, use_loss=True, rank_number=10, space_type='\t'):
        # external vars
        self.train_set = ReadFile(train_file, space_type=space_type).return_information()
        self.test_file = test_file
        self.ranking_file = ranking_file
        self.factors = factors
        self.learn_rate = learn_rate
        self.predict_items_number = predict_items_number
        self.init_mean = init_mean
        self.init_stdev = init_stdev
        self.num_interactions = num_interactions
        self.reg_bias = reg_bias
        self.reg_u = reg_u
        self.reg_i = reg_i
        self.reg_j = reg_j
        self.use_loss = use_loss
        self.rank_number = rank_number
        self.train_set["users"] = list(self.train_set['users'])
        self.train_set["items"] = list(self.train_set['items'])
        self.users = list(self.train_set['users'])
        self.items = list(self.train_set['items'])
        if self.test_file is not None:
            self.test_set = ReadFile(test_file).return_information()
            self.test_set['users'] = list(self.test_set['users'])
            self.test_set['items'] = list(self.test_set['items'])
            self.users = sorted(self.train_set['users'] + self.test_set['users'])
            self.items = sorted(self.train_set['items'] + self.test_set['items'])
        if num_events is None:
            self.num_events = self.train_set['ni']
        else:
            self.num_events = num_events

        # internal vars
        self.loss = None
        self.loss_sample = list()
        self.ranking = list()
        self.map_items = dict()
        self.map_items_index = dict()
        self.map_users = dict()
        self.map_users_index = dict()

        for i, item in enumerate(self.items):
            self.map_items.update({item: i})
            self.map_items_index.update({i: item})
        for u, user in enumerate(self.users):
            self.map_users.update({user: u})
            self.map_users_index.update({u: user})

    def _create_factors(self):
        self.p = np.random.normal(self.init_mean, self.init_stdev, (len(self.users), self.factors))
        self.q = np.random.normal(self.init_mean, self.init_stdev, (len(self.items), self.factors))
        # self.bias = self.init_mean * np.random.randn(self.number_items, 1) + self.init_stdev ** 2
        self.bias = np.zeros(len(self.items), np.double)

    def _sample_triple(self):
        user = np.random.choice(self.train_set["users"])
        item_i = np.random.choice(list(self.train_set['feedback'][user].keys()))
        item_j = np.random.choice(self.train_set['not_seen'][user])
        return self.map_users[user], self.map_items[item_i], self.map_items[item_j]

    def _predict(self, user, item):
        return self.bias[item] + np.dot(self.p[user], self.q[item])

    def _update_factors(self, user, item_i, item_j):
        # Compute Difference
        eps = 1 / (1 + np.exp(self._predict(user, item_i) - self._predict(user, item_j)))

        self.bias[item_i] += self.learn_rate * (eps - self.reg_bias * self.bias[item_i])
        self.bias[item_j] += self.learn_rate * (eps - self.reg_bias * self.bias[item_j])

        # Adjust the factors
        u_f = self.p[user]
        i_f = self.q[item_i]
        j_f = self.q[item_j]

        # Compute and apply factor updates
        self.p[user] += self.learn_rate * ((i_f - j_f) * eps - self.reg_u * u_f)
        self.q[item_i] += self.learn_rate * (u_f * eps - self.reg_i * i_f)
        self.q[item_j] += self.learn_rate * (-u_f * eps - self.reg_j * j_f)

    def _compute_loss(self):
        ranking_loss = 0
        for sample in self.loss_sample:
            x_uij = self._predict(sample[0], sample[1]) - self._predict(sample[0], sample[2])
            ranking_loss += 1 / (1 + np.exp(x_uij))

        complexity = 0
        for sample in self.loss_sample:
            complexity += self.reg_u * np.linalg.norm(self.p[sample[0]]) ** 2
            complexity += self.reg_i * np.linalg.norm(self.q[sample[1]]) ** 2
            complexity += self.reg_j * np.linalg.norm(self.q[sample[2]]) ** 2
            complexity += self.reg_bias * self.bias[sample[1]] ** 2
            complexity += self.reg_bias * self.bias[sample[2]] ** 2

        return ranking_loss + 0.5 * complexity

    # Perform one iteration of stochastic gradient ascent over the training data
    # One iteration is samples number of positive entries in the training matrix times
    def train_model(self):
        if self.use_loss:
            num_sample_triples = int(np.sqrt(len(self.users)) * 100)
            for _ in range(num_sample_triples):
                self.loss_sample.append(self._sample_triple())
            self.loss = self._compute_loss()

        for n in range(self.num_interactions):
            for _ in range(self.num_events):
                u, i, j = self._sample_triple()
                self._update_factors(u, i, j)

            if self.use_loss:
                actual_loss = self._compute_loss()
                if actual_loss > self.loss:
                    self.learn_rate *= 0.5
                elif actual_loss < self.loss:
                    self.learn_rate *= 1.1
                self.loss = actual_loss

            print("step::", n, "RMSE::", self.loss)

    def predict(self):
        w = self.bias.T + np.dot(self.p, self.q.T)
        for u, user in enumerate(self.users):
            partial_ranking = list()
            user_list = sorted(range(len(w[u])), key=lambda k: w[u][k], reverse=True)
            for i in user_list[:100]:
                item = self.map_items_index[i]
                try:
                    if item not in self.train_set["feedback"][user]:
                        partial_ranking.append((user, item, w[u][i]))
                except KeyError:
                    partial_ranking.append((user, item, w[u][i]))
            self.ranking += partial_ranking[:self.rank_number]

        if self.ranking_file is not None:
            self.ranking = sorted(self.ranking, key=lambda x: x[0])
            WriteFile(self.ranking_file, self.ranking).write_recommendation()

    def evaluate(self, measures):
        res = ItemRecommendationEvaluation().evaluation_ranking(self.ranking, self.test_file)
        evaluation = 'Eval:: '
        for measure in measures:
            evaluation += measure + ': ' + str(res[measure]) + ' '
        print(evaluation)

    def execute(self, measures=('Prec@5', 'Prec@10', 'NDCG@5', 'NDCG@10', 'MAP@5', 'MAP@10')):
        # methods
        print("[Case Recommender: Item Recommendation > BPR MF Algorithm]\n")
        print("training data:: ", len(self.train_set['users']), " users and ", len(self.train_set['items']),
              " items and ", self.train_set['ni'], " interactions | sparsity ", self.train_set['sparsity'])

        if self.test_file is not None:
            test_set = ReadFile(self.test_file).return_information()
            print("test data:: ", len(self.test_set['users']), " users and ", len(self.test_set['items']),
                  " items and ", (self.test_set['ni']), " interactions | sparsity ", self.test_set['sparsity'])
            del test_set
        self._create_factors()
        print("training time:: ", timed(self.train_model), " sec")
        print("prediction_time:: ", timed(self.predict), " sec\n")
        if self.test_file is not None:
            self.evaluate(measures)



# coding=utf-8
"""
© 2016. Case Recommender All Rights Reserved (License GPL3)

User Based Collaborative Filtering Recommender

    User-kNN predicts a user’s rating according to how similar users rated the same item. The algorithm matches similar
    users based on the similarity of their ratings on items.

    Literature:
        http://files.grouplens.org/papers/algs.pdf

Parameters
-----------
    - train_file: string
    - test_file: string
    - prediction_file: string
        file to write final prediction
    - similarity_metric: string
        Pairwise metric to compute the similarity between the users.
        Reference about distances:
            - http://docs.scipy.org/doc/scipy-0.14.0/reference/generated/scipy.spatial.distance.pdist.html
    - neighbors: int
        The number of user candidates strategy that you can choose for selecting the possible items to recommend.
    - space_type: string

"""

import numpy as np
from scipy.spatial.distance import squareform, pdist
from caserec.utils.extra_functions import timed
from caserec.utils.read_file import ReadFile
from caserec.recommenders.rating_prediction.base_knn import BaseKNNRecommenders
from caserec.utils.write_file import WriteFile

__author__ = 'Arthur Fortes'


class UserKNN(BaseKNNRecommenders):
    def __init__(self, train_file, test_file, prediction_file=None, similarity_metric="correlation", neighbors=30,
                 space_type='\t'):
        self.train_set = ReadFile(train_file, space_type=space_type).return_information()
        self.test_set = ReadFile(test_file, space_type=space_type).return_information()
        BaseKNNRecommenders.__init__(self, self.train_set, self.test_set)
        self.k = neighbors
        self.similarity_metric = similarity_metric
        self.prediction_file = prediction_file
        self.predictions = list()
        self.su_matrix = None

    def compute_similarity(self):
        # Calculate distance matrix between users
        self.su_matrix = np.float32(squareform(pdist(self.matrix, self.similarity_metric)))
        # transform distances in similarities
        self.su_matrix = 1 - self.su_matrix
        del self.matrix

    """
    for each pair (u,i) in test set, this method returns a prediction based
    on the others feedback in the train set

    rui = bui + (sum((rvi - bvi) * sim(u,v)) / sum(sim(u,v)))

    """

    def predict(self):
        if self.test is not None:
            for user in self.test['users']:
                for item in self.test['feedback'][user]:
                    list_n = list()
                    try:
                        ruj = 0.0
                        sum_sim = 0.0

                        for user_j in self.train['di'][item]:
                            sim = self.su_matrix[self.map_users[user]][self.map_users[user_j]]
                            if np.math.isnan(sim):
                                sim = 0.0
                            list_n.append((user_j, sim))
                        list_n = sorted(list_n, key=lambda x: -x[1])

                        for pair in list_n[:self.k]:
                            try:
                                ruj += (self.train_set['feedback'][pair[0]][item] -
                                        self.bui[pair[0]][item]) * pair[1]
                                sum_sim += pair[1]
                            except KeyError:
                                pass

                        try:
                            ruj = self.bui[user][item] + (ruj / sum_sim)
                        except ZeroDivisionError:
                            ruj = self.bui[user][item]

                    except KeyError:
                        ruj = self.bui[user][item]

                    # normalize the ratings based on the highest and lowest value.
                    if ruj > self.train_set["max"]:
                        ruj = self.train_set["max"]
                    if ruj < self.train_set["min"]:
                        ruj = self.train_set["min"]

                    self.predictions.append((user, item, ruj))

            if self.prediction_file is not None:
                WriteFile(self.prediction_file, self.predictions).write_recommendation()
            return self.predictions

    def execute(self):
        # methods
        print("[Case Recommender: Rating Prediction > User Algorithm]\n")
        print("training data:: ", len(self.train_set['users']), " users and ", len(self.train_set['items']),
              " items and ", self.train_set['ni'], " interactions | sparsity ", self.train_set['sparsity'])
        print("test data:: ", len(self.test_set['users']), " users and ", len(self.test_set['items']),
              " items and ", (self.test_set['ni']), " interactions | sparsity ", self.test_set['sparsity'])
        # training baselines bui
        self.fill_matrix()
        print("training time:: ", timed(self.train_baselines), " sec")
        self.compute_similarity()
        print("prediction_time:: ", timed(self.predict), " sec\n")
        self.evaluate(self.predictions)



# coding=utf-8
"""
© 2016. Case Recommender All Rights Reserved (License GPL3)

User Based Collaborative Filtering Recommender with Attributes

    User-Attribute-kNN predicts a user’s rating according to how similar users rated the same item. The algorithm
    matches similar users based on the similarity of their attributes scores. However, instead of traditional UserKNN,
    this approach uses a pre-computed similarity matrix.

    Literature:
        More details: http://files.grouplens.org/papers/algs.pdf

Parameters
-----------
    - train_file: string
    - test_file: string
    - prediction_file: string
        file to write final prediction
    - metadata_file: string
        Metadata file ; Format file:
        item \t metadata \t value\n
    - similarity_matrix_file: string
        Pairwise metric to compute the similarity between the users based on a set of attributes.
        Format file:
        Distances separated by \t, where the users should be ordering. E g.:
        distance1\tdistance2\tdistance3\n
        distance1\tdistance2\tdistance3\n
        distance1\tdistance2\tdistance3\n
    - neighbors: int
        The number of user candidates strategy that you can choose for selecting the possible items to recommend.
    - similarity_metric: string
    - space_type: string

"""

import sys
from caserec.utils.extra_functions import timed
from caserec.utils.read_file import ReadFile
from caserec.recommenders.rating_prediction.userknn import UserKNN

__author__ = 'Arthur Fortes'


class UserAttributeKNN(UserKNN):
    def __init__(self, train_file, test_file, metadata_file=None, similarity_matrix_file=None, prediction_file=None,
                 neighbors=30, similarity_metric="correlation", space_type='\t'):
        UserKNN.__init__(self, train_file, test_file, prediction_file=prediction_file, neighbors=neighbors,
                         similarity_metric=similarity_metric, space_type=space_type)

        if metadata_file is None and similarity_matrix_file is None:
            print("This algorithm needs a similarity matrix or a metadata file!")
            sys.exit(0)

        if metadata_file is not None:
            self.metadata = ReadFile(metadata_file, space_type).read_metadata(self.users)
            self.matrix = self.metadata['matrix']
        self.similarity_matrix_file = similarity_matrix_file

    def read_matrix(self):
        self.su_matrix = ReadFile(self.similarity_matrix_file).read_matrix()

    def execute(self):
        # methods
        print("[Case Recommender: Rating Prediction > User Attribute KNN Algorithm]\n")
        print("training data:: ", len(self.train_set['users']), " users and ", len(
            self.train_set['items']), " items and ", self.train_set['ni'], " interactions")
        print("test data:: ", len(self.test_set['users']), " users and ", len(self.test_set['items']),
              " items and ", self.test_set['ni'], " interactions")

        # training baselines bui
        print("training time:: ", timed(self.train_baselines), " sec")
        if self.similarity_matrix_file is not None:
            print("compute similarity:: ", timed(self.read_matrix), " sec")
        else:
            print("compute similarity time:: ", timed(self.compute_similarity), " sec")
        print("prediction_time:: ", timed(self.predict), " sec\n")
        self.evaluate(self.predictions)



# coding=utf-8
"""
© 2016. Case Recommender All Rights Reserved (License GPL3)

Ensemble Average Based
    Literature:
        Improving Personalized Ranking in Recommender Systems with Multimodal Interactions.
        WIC 2014.
        http://dx.doi.org/10.1109/wi-iat.2014.34

Parameters
-----------

"""
from caserec.evaluation.item_recommendation import ItemRecommendationEvaluation
from caserec.utils.read_file import ReadFile
from caserec.utils.write_file import WriteFile
from caserec.utils.extra_functions import check_len_lists

__author__ = 'Arthur Fortes'


# utils
def return_list_info(train_set):
    lu = set()
    li = set()
    dict_users_interactions = dict()
    dict_non_seen_items = dict()
    dict_index = dict()

    for interaction in train_set:
        user, item, score = interaction[0], interaction[1], interaction[2]
        lu.add(user)
        li.add(item)
        dict_users_interactions.setdefault(user, {}).update({item: score})

    for u, user in enumerate(lu):
        dict_index.setdefault('users', {}).update({user: u})
        dict_non_seen_items[user] = list(li - set(dict_users_interactions[user].keys()))

    for i, item in enumerate(li):
        dict_index.setdefault('items', {}).update({item: i})

    return dict_users_interactions, dict_non_seen_items, lu, li, dict_index


class EnsembleAverageBased(object):
    def __init__(self, list_train_files, list_rank_files, file_write, test_file=None, rank_number=10, space_type='\t'):
        self.list_train_files = list_train_files
        self.list_rank_files = list_rank_files
        self.file_write = file_write
        self.test_file = test_file
        self.rank_number = rank_number
        self.space_type = space_type
        check_len_lists(self.list_train_files, self.list_rank_files)
        self.num_interactions = len(self.list_train_files)
        self.factors = list()
        self.individual_datasets = list()
        self.final_dataset = list()
        self.betas = list()

        # vars
        self.dict_item = dict()
        self.dict_not_item = dict()
        self.list_users = set()
        self.list_items = set()
        self.dict_index = dict()
        self.rankings = list()
        self.final_ranking = list()
        self.normalization = list()
        self.averages = list()
        self.num_user_int = list()

    def read_ranking_files(self):
        for ranking_file in self.list_rank_files:
            ranking = ReadFile(ranking_file, space_type=self.space_type)
            rank_interaction, list_interaction = ranking.read_rankings()
            self.rankings.append(rank_interaction)
            self.normalization.append([min(list_interaction), max(list_interaction)])
            self.averages.append(ranking.average_scores)
            self.num_user_int.append(ranking.num_user_interactions)

    def treat_interactions(self):
        for num, interaction_file in enumerate(self.list_train_files):
            interaction = ReadFile(interaction_file, space_type=self.space_type)
            interaction.triple_information()
            self.individual_datasets.append(interaction.triple_dataset)
            self.final_dataset += interaction.triple_dataset

        self.dict_item, self.dict_not_item, self.list_users, self.list_items, \
            self.dict_index = return_list_info(self.final_dataset)

        self.list_users = list(self.list_users)
        self.list_items = list(self.list_items)

    def ensemble_ranks(self):
        for u, user in enumerate(self.list_users):
            list_items = list()
            for item in self.dict_not_item[user]:
                scores = list()
                control = 0
                for m in range(self.num_interactions):
                    try:
                        score = self.rankings[m][user].get(item, 0)
                        if score > 0:
                            average = float(self.averages[m][user]) / float(self.num_user_int[m][user])
                            if score >= average:
                                control += 1

                            score = (score - self.normalization[m][0]) / (
                                self.normalization[m][1] - self.normalization[m][0])
                            scores.append(score)

                    except KeyError:
                        pass

                if len(scores) == self.num_interactions:

                    if control == self.num_interactions:
                        rui = (float(sum(scores)) / float(len(scores)))
                    else:
                        rui = max(scores)
                    list_items.append([item, rui])

            list_items = sorted(list_items, key=lambda x: -x[1])
            self.final_ranking.append([user, list_items[:self.rank_number]])

    def write_ranking(self):
        self.final_ranking = sorted(self.final_ranking, key=lambda x: x[0])
        write_ensemble = WriteFile(self.file_write, self.final_ranking, self.space_type)
        write_ensemble.write_recommendation()

    def evaluate(self, measures):
        res = ItemRecommendationEvaluation().evaluation_ranking(self.final_ranking, self.test_file)
        evaluation = 'Eval:: '
        for measure in measures:
            evaluation += measure + ': ' + str(res[measure]) + ' '
        print(evaluation)

    def execute(self, measures=('Prec@5', 'Prec@10', 'NDCG@5', 'NDCG@10', 'MAP@5', 'MAP@10')):
        # methods
        print("[Case Recommender: Item Recommendation > Ensemble Average Based Algorithm]\n")
        # call internal methods
        self.read_ranking_files()
        print("Read Ranking Files...")
        self.treat_interactions()
        self.ensemble_ranks()
        print("Finished Ensemble interactions...")
        self.write_ranking()

        if self.test_file is not None:
            self.evaluate(measures)



# coding=utf-8
"""
© 2016. Case Recommender All Rights Reserved (License GPL3)

User Based Collaborative Filtering Recommender with Attributes

    User-Attribute-kNN predicts a user’s rating according to how similar users rated the same item. The algorithm
    matches similar users based on the similarity of their attributes scores. However, instead of traditional UserKNN,
    this approach uses a pre-computed similarity matrix.

    Literature:
        More details: http://files.grouplens.org/papers/algs.pdf

Parameters
-----------
    - train_file: string
    - test_file: string
    - metadata_file: string
        Metadata file ; Format file:
        item \t metadata \t value\n
    - similarity_matrix_file: file
        Pairwise metric to compute the similarity between the users based on a set of attributes.
        Format file:
        Distances separated by \t, where the users should be ordering. E g.:
        distance1\tdistance2\tdistance3\n
        distance1\tdistance2\tdistance3\n
        distance1\tdistance2\tdistance3\n
    - ranking_file: string
        file to write final ranking
    - neighbors: int
        The number of user candidates strategy that you can choose for selecting the possible items to recommend.
    - similarity_metric: string
        Pairwise metric to compute the similarity between the users.
        Reference about distances:
            - http://docs.scipy.org/doc/scipy-0.14.0/reference/generated/scipy.spatial.distance.pdist.html
    - space_type: string
"""

import sys
from caserec.recommenders.item_recommendation.userknn import UserKNN
from caserec.utils.extra_functions import timed
from caserec.utils.read_file import ReadFile

__author__ = "Arthur Fortes"


class UserAttributeKNN(UserKNN):
    def __init__(self, train_file, test_file=None, metadata_file=None, similarity_matrix_file=None, ranking_file=None,
                 neighbors=30, rank_number=10, similarity_metric="correlation", space_type='\t'):
        UserKNN.__init__(self, train_file, test_file=test_file, ranking_file=ranking_file, neighbors=neighbors,
                         rank_number=rank_number, similarity_metric=similarity_metric, space_type=space_type)

        if metadata_file is None and similarity_matrix_file is None:
            print("This algorithm needs a similarity matrix or a metadata file!")
            sys.exit(0)

        if metadata_file is not None:
            self.metadata = ReadFile(metadata_file, space_type=space_type).read_metadata(self.users)
            self.matrix = self.metadata['matrix']
        self.similarity_matrix_file = similarity_matrix_file

    def read_matrix(self):
        self.su_matrix = ReadFile(self.similarity_matrix_file).read_matrix()

    def execute(self, measures=('Prec@5', 'Prec@10', 'NDCG@5', 'NDCG@10', 'MAP@5', 'MAP@10')):
        print("[Case Recommender: Item Recommendation > User Attribute KNN Algorithm]\n")
        print("training data:: ", len(self.train_set['users']), " users and ", len(self.train_set['items']),
              " items and ", self.train_set['ni'], " interactions | sparsity ", self.train_set['sparsity'])

        if self.test_file is not None:
            test_set = ReadFile(self.test_file, space_type=self.space_type).return_information()
            print("test data:: ", len(self.test_set['users']), " users and ", len(self.test_set['items']),
                  " items and ", (self.test_set['ni']), " interactions | sparsity ", self.test_set['sparsity'])
            del test_set

        if self.similarity_matrix_file is not None:
            print("training time:: ", timed(self.read_matrix), " sec")
        else:
            print("training time:: ", timed(self.compute_similarity), " sec")
            print("prediction_time:: ", timed(self.predict), " sec\n")
        if self.test_file is not None:
            self.evaluate(measures)
